######################################################################################################################

# -- Variational Inference for Gravitational wave Parameter Estimation --


#######################################################################################################################

import warnings
warnings.filterwarnings("ignore")
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3' 
import argparse
import numpy as np
import tensorflow as tf
from tensorflow.python.util import deprecation
deprecation._PRINT_DEPRECATION_WARNINGS = False
import scipy.io as sio
import h5py
import sys
from sys import exit
import shutil
import bilby
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import time
from time import strftime
import corner
import glob
from matplotlib.lines import Line2D
import pandas as pd
import logging.config
from contextlib import contextmanager
import json
from lal import GreenwichMeanSiderealTime

import skopt
from skopt import gp_minimize, forest_minimize
from skopt.space import Real, Categorical, Integer
from skopt.plots import plot_convergence
from skopt.plots import plot_objective, plot_evaluations
from skopt.utils import use_named_args

try:
    from .models import CVAE_model
    from .gen_benchmark_pe import run
    from . import plotting
    from .plotting import prune_samples
except ModuleNotFoundError:
    from models import CVAE_model
    from gen_benchmark_pe import run
    import plotting
    from plotting import prune_samples

# Check for optional basemap installation
try:
    from mpl_toolkits.basemap import Basemap
    print("module 'basemap' is installed")
except (ModuleNotFoundError, ImportError):
    print("module 'basemap' is not installed")
    print("Skyplotting functionality is automatically disabled.")
    skyplotting_usage = False
else:
    skyplotting_usage = True
    try:
        from .skyplotting import plot_sky
    except:
        from skyplotting import plot_sky

""" Script has 4 main functions:
1.) Generate training data
2.) Generate testing data
3.) Train model
4.) Test model
5.) Generate samples only given model and timeseries
"""

parser = argparse.ArgumentParser(description='VItamin: A user friendly Bayesian inference machine learning library.')
parser.add_argument("--gen_train", default=False, help="generate the training data")
parser.add_argument("--gen_test", default=False, help="generate the testing data")
parser.add_argument("--train", default=False, help="train the network")
parser.add_argument("--resume_training", default=False, help="resume training of network")
parser.add_argument("--test", default=False, help="test the network")
parser.add_argument("--params_file", default=None, type=str, help="dictionary containing parameters of run")
parser.add_argument("--bounds_file", default=None, type=str, help="dictionary containing source parameter bounds")
parser.add_argument("--fixed_vals_file", default=None, type=str, help="dictionary containing source parameter values when fixed")
parser.add_argument("--pretrained_loc", default=None, type=str, help="location of a pretrained network (i.e. .ckpt file)")
parser.add_argument("--test_set_loc", default=None, type=str, help="directory containing test set waveforms")
parser.add_argument("--gen_samples", default=False, help="If True, generate samples only (no plotting)")
parser.add_argument("--num_samples", type=int, default=10000, help="number of posterior samples to generate")
parser.add_argument("--use_gpu", default=False, help="if True, use gpu")
args = parser.parse_args()

global params; global bounds; global fixed_vals

# Define default location of the parameters files
params = os.path.join(os.getcwd(), 'params_files', 'params.json')
bounds = os.path.join(os.getcwd(), 'params_files', 'bounds.json')
fixed_vals = os.path.join(os.getcwd(), 'params_files', 'fixed_vals.json')

# Load parameters files
if args.params_file != None:
    params = args.params_file
if args.bounds_file != None:
    bounds = args.bounds_file
if args.fixed_vals_file != None:
    fixed_vals = args.fixed_vals_file

# Ranges over which hyperparameter optimization parameters are allowed to vary
kernel_1 = Integer(low=3, high=12, name='kernel_1')
strides_1 = Integer(low=1, high=2, name='strides_1')
pool_1 = Integer(low=1, high=2, name='pool_1')
kernel_2 = Integer(low=3, high=12, name='kernel_2')
strides_2 = Integer(low=1, high=2, name='strides_2')
pool_2 = Integer(low=1, high=2, name='pool_2')
kernel_3 = Integer(low=3, high=12, name='kernel_3')
strides_3 = Integer(low=1, high=2, name='strides_3')
pool_3 = Integer(low=1, high=2, name='pool_3')
kernel_4 = Integer(low=3, high=12, name='kernel_4')
strides_4 = Integer(low=1, high=2, name='strides_4')
pool_4 = Integer(low=1, high=2, name='pool_4')

z_dimension = Integer(low=7, high=100, name='z_dimension')
n_modes = Integer(low=7, high=12, name='n_modes')
n_filters_1 = Integer(low=32, high=33, name='n_filters_1')
n_filters_2 = Integer(low=32, high=33, name='n_filters_2')
n_filters_3 = Integer(low=32, high=33, name='n_filters_3')
n_filters_4 = Integer(low=32, high=33, name='n_filters_4')
batch_size = Integer(low=511, high=512, name='batch_size')
n_weights_fc_1 = Integer(low=2047, high=2048, name='n_weights_fc_1')
n_weights_fc_2 = Integer(low=2047, high=2048, name='n_weights_fc_2')
n_weights_fc_3 = Integer(low=2047, high=2048, name='n_weights_fc_3')

# putting defined hyperparameter optimization ranges into a list
dimensions = [kernel_1, 
              strides_1,
              pool_1,
              kernel_2, 
              strides_2,
              pool_2,
              kernel_3,
              strides_3,
              pool_3,
              kernel_4,
              strides_4,
              pool_4,
              z_dimension,
              n_modes,
              n_filters_1,
              n_filters_2,
              n_filters_3,
              n_filters_4,
              batch_size,
              n_weights_fc_1,
              n_weights_fc_2,
              n_weights_fc_3]

"""
# list of initial default hyperparameters to use for GP hyperparameter optimization
default_hyperparams = [params['filter_size_r1'][0],
                       params['conv_strides_r1'][0],
                       params['maxpool_r1'][0],
                       params['filter_size_r1'][1],
                       params['conv_strides_r1'][1],
                       params['maxpool_r1'][1],
                       params['filter_size_r1'][2],
                       params['conv_strides_r1'][2],
                       params['maxpool_r1'][2],
                       params['filter_size_r1'][3],
                       params['conv_strides_r1'][3],
                       params['maxpool_r1'][3],
                       params['z_dimension'],
                       params['n_modes'],
                       params['n_filters_r1'][0],
                       params['n_filters_r1'][1],
                       params['n_filters_r1'][2],
                       params['n_filters_r1'][3],
                       params['batch_size'],
                       params['n_weights_r1'][0],
                       params['n_weights_r1'][1],
                       params['n_weights_r1'][2],
                      ]
"""

# dummy value for initial hyperparameter best KL (to be minimized). Doesn't need to be changed.
best_loss = int(1e6)

@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout

def load_data(params,bounds,fixed_vals,input_dir,inf_pars,load_condor=False):
    """ Function to load either training or testing data.

    Parameters
    ----------
    params: dict
        Dictionary containing parameter values of run
    bounds: dict
        Dictionary containing the allowed bounds of source GW parameters
    fixed_vals: dict
        Dictionary containing the fixed values of GW source parameters
    input_dir: str
        Directory where training or testing files are stored
    inf_pars: list
        list of parameters to infer when training ML model
    load_condor: bool
        if True, load test samples rather than training samples

    Returns
    -------
    x_data: array_like
        array containing training/testing source parameter values
    y_data: array_like
        array containing training/testing noise-free times series
    y_data_noisy: array_like
        array containing training/testing noisy time series	
    y_normscale: float
        value by which to normalize all time series to be between zero and one
    snrs: array_like
        array containing optimal snr values for all training/testing time series
    """

    train_files = []
    
    # Get list of all training/testing files and define dictionary to store values in files
    if type("%s" % input_dir) is str:
        dataLocations = ["%s" % input_dir]
        data={'x_data': [], 'y_data_noisefree': [], 'y_data_noisy': [], 'rand_pars': []}

    # Sort files from first generated to last generated
    if load_condor == True:
        filenames = sorted(os.listdir(dataLocations[0]), key=lambda x: int(x.split('.')[0].split('_')[-1]))
    else:
        filenames = os.listdir(dataLocations[0])

    # Append training/testing filenames to list. Ignore those that can't be loaded
    snrs = []
    for filename in filenames:
        try:
            train_files.append(filename)

        except OSError:
            print('Could not load requested file')
            continue

    # If loading by chunks, randomly shuffle list of training/testing filenames
    if params['load_by_chunks'] == True and load_condor == False:
        train_files_idx = np.arange(len(train_files))[:int(params['load_chunk_size']/1000.0)]
        np.random.shuffle(train_files)
        train_files = np.array(train_files)[train_files_idx]

    # Iterate over all training/testing files and store source parameters, time series and SNR info in dictionary
    for filename in train_files:
        try:
            data_temp={'x_data': h5py.File(dataLocations[0]+'/'+filename, 'r')['x_data'][:],
                  'y_data_noisefree': h5py.File(dataLocations[0]+'/'+filename, 'r')['y_data_noisefree'][:],
                  'y_data_noisy': h5py.File(dataLocations[0]+'/'+filename, 'r')['y_data_noisy'][:],
                  'rand_pars': h5py.File(dataLocations[0]+'/'+filename, 'r')['rand_pars'][:]}
            data['y_data_noisefree'].append(np.expand_dims(data_temp['y_data_noisefree'], axis=0))
            snrs.append(h5py.File(dataLocations[0]+'/'+filename, 'r')['snrs'][:])
            data['x_data'].append(data_temp['x_data'])
            data['y_data_noisy'].append(np.expand_dims(data_temp['y_data_noisy'], axis=0))
            data['rand_pars'] = data_temp['rand_pars']
            print('... Loaded file ' + filename)
        except OSError:
            print('Could not load requested file')
            continue
    snrs = np.array(snrs)

    # Extract the prior bounds from training/testing files
    data['x_data'] = np.concatenate(np.array(data['x_data']), axis=0).squeeze()
    data['y_data_noisefree'] = np.concatenate(np.array(data['y_data_noisefree']), axis=0)
    data['y_data_noisy'] = np.concatenate(np.array(data['y_data_noisy']), axis=0)
    
    # expand dimensions if only using one test sample
    if data['x_data'].ndim == 1:
        data['x_data'] = np.expand_dims(data['x_data'],axis=0)

    # only convert to sidereal for training data    
    if load_condor==False:
        # get geocenttime and ra index
        for i,k in enumerate(data_temp['rand_pars']):
            k = k.decode('utf-8')
            if k == 'geocent_time':
                geo_idx = i
            elif k == 'ra':
                ra_idx = i 
        # Check if both geocentime and RA exist
        try:
            geo_idx; ra_idx
        except NameError:
            print('Either time or RA is fixed. Not converting RA to hour angle.')
        else:
            # Iterate over all training samples and convert to hour angle
            for i in range(data['x_data'].shape[0]):
#                data['x_data'][i,ra_idx]=np.mod(GreenwichMeanSiderealTime(float(params['ref_geocent_time']+data['x_data'][i,geo_idx])) - data['x_data'][i,ra_idx], 2.0*np.pi)
                data['x_data'][i,ra_idx]=np.mod(GreenwichMeanSiderealTime(params['ref_geocent_time']) - data['x_data'][i,ra_idx], 2.0*np.pi)

    # Normalise the source parameters np.remainder(blah,np.pi)
    for i,k in enumerate(data_temp['rand_pars']):
        par_min = k.decode('utf-8') + '_min'
        par_max = k.decode('utf-8') + '_max'

        # ensure psi is 0 to pi
        if par_min == 'psi_min':
            data['x_data'][:,i]=np.remainder(data['x_data'][:,i],np.pi)

        # normalize by bounds
        data['x_data'][:,i]=(data['x_data'][:,i] - bounds[par_min]) / (bounds[par_max] - bounds[par_min])

    x_data = data['x_data']
    y_data = data['y_data_noisefree']
    y_data_noisy = data['y_data_noisy']

    # Define time series normalization factor to use on test samples. We consistantly use the same normscale value if loading by chunks
    y_normscale = params['y_normscale']   
 
    # extract inference parameters from all source parameters loaded earlier
    idx = []
    print()
    for k in inf_pars:
        if load_condor == False:
            print('... ' + k + ' will be inferred')
        for i,q in enumerate(data['rand_pars']):
            m = q.decode('utf-8')
            if k==m:
                idx.append(i)
    x_data = x_data[:,idx]

    return x_data, y_data, y_data_noisy, y_normscale, snrs

@use_named_args(dimensions=dimensions)
def hyperparam_fitness(kernel_1, strides_1, pool_1,
                       kernel_2, strides_2, pool_2,
                       kernel_3, strides_3, pool_3,
                       kernel_4, strides_4, pool_4,
                       z_dimension,n_modes,
                       n_filters_1,n_filters_2,n_filters_3,n_filters_4,
                       batch_size,
                       n_weights_fc_1,n_weights_fc_2,n_weights_fc_3):
    """ Fitness function used in Gaussian Process hyperparameter optimization 
    Returns a value to be minimized (in this case, the total loss of the 
    neural network during training.

    Parameters
    ----------
    kernel_1: skopt function
        Range over which kernel size in first CNN layer is allowed to vary
    kernel_2: skopt function
        Range over which kernel size in second CNN layer is allowed to vary
    kernel_3: skopt function
        Range over which kernel size in third CNN layer is allowed to vary  
    kernel_4: skopt function
        Range over which kernel size in fourth CNN layer is allowed to vary
    strides_1: skopt function
        Range over which stride size in first CNN layer is allowed to vary 
    strides_2: skopt function
        Range over which stride size in second CNN layer is allowed to vary
    strides_3: skopt function
        Range over which stride size in third CNN layer is allowed to vary
    strides_4: skopt function
        Range over which stride size in fourth CNN layer is allowed to vary
    pool_1: skopt function
        Range over which pool size in first CNN layer is allowed to vary
    pool_2: skopt function
        Range over which pool size in second CNN layer is allowed to vary
    pool_3: skopt function
        Range over which pool size in third CNN layer is allowed to vary
    pool_4: skopt function
        Range over which pool size in fourth CNN layer is allowed to vary
    z_dimension: skopt function
        Range over which latent space is allowed to vary
    n_modes: skopt function
        Range over which the number of gaussian modes in latent space is allowed to vary
    n_filters_1: skopt function
        Range over which the number of filters in first CNN layer is allowed to vary
    n_filters_2: skopt function
        Range over which the number of filters in second CNN layer is allowed to vary
    n_filters_3: skopt function
        Range over which the number of filters in third CNN layer is allowed to vary
    n_filters_4: skopt function
        Range over which the number of filters in fourth CNN layer is allowed to vary
    batch_size: skopt function
        Range over which the batch size is allowed to vary
    n_weights_fc_1: skopt function
        Range over which the number of neurons in the first hidden layer is allowed to vary
    n_weights_fc_2: skopt function
        Range over which the number of neurons in the second hidden layer is allowed to vary
    n_weights_fc_3: skopt function
        Range over which the number of neurons in the third hidden layer is allowed to vary

    Returns
    -------
    VICI_loss: float
        Total loss of the current optimized network
    """

    # set tunable hyper-parameters
    params['filter_size_r1'] = [kernel_1,kernel_2,kernel_3,kernel_4]
    params['filter_size_r2'] = [kernel_1,kernel_2,kernel_3,kernel_4]
    params['filter_size_q'] = [kernel_1,kernel_2,kernel_3,kernel_4]
    params['n_filters_r1'] = [n_filters_1,n_filters_2,n_filters_3,n_filters_4]
    params['n_filters_r2'] = [n_filters_1,n_filters_2,n_filters_3,n_filters_4]
    params['n_filters_q'] = [n_filters_1,n_filters_2,n_filters_3,n_filters_4]

    # number of filters has to be odd for some reason (this ensures that this is the case)
    for filt_idx in range(len(params['n_filters_r1'])):
        if (params['n_filters_r1'][filt_idx] % 3) != 0:

            # keep adding 1 until filter size is divisible by 3
            while (params['n_filters_r1'][filt_idx] % 3) != 0:
                params['n_filters_r1'][filt_idx] += 1
                params['n_filters_r2'][filt_idx] += 1
                params['n_filters_q'][filt_idx] += 1
    params['conv_strides_r1'] = [strides_1,strides_2,strides_3,strides_4]
    params['conv_strides_r2'] = [strides_1,strides_2,strides_3,strides_4] 
    params['conv_strides_q'] = [strides_1,strides_2,strides_3,strides_4] 
    params['maxpool_r1'] = [pool_1,pool_2,pool_3,pool_4]
    params['maxpool_r2'] = [pool_1,pool_2,pool_3,pool_4]
    params['maxpool_q'] = [pool_1,pool_2,pool_3,pool_4]
    params['pool_strides_r1'] = [pool_1,pool_2,pool_3,pool_4]
    params['pool_strides_r2'] = [pool_1,pool_2,pool_3,pool_4]
    params['pool_strides_q'] = [pool_1,pool_2,pool_3,pool_4]
    params['z_dimension'] = z_dimension
    params['n_modes'] = n_modes
    params['batch_size'] = batch_size
    params['n_weights_r1'] = [n_weights_fc_1,n_weights_fc_2,n_weights_fc_3]
    params['n_weights_r2'] = [n_weights_fc_1,n_weights_fc_2,n_weights_fc_3]
    params['n_weights_q'] = [n_weights_fc_1,n_weights_fc_2,n_weights_fc_3]

    # Print the hyper-parameters.
    print('kernel_1: {}'.format(kernel_1))
    print('strides_1: {}'.format(strides_1))
    print('pool_1: {}'.format(pool_1))
    print('kernel_2: {}'.format(kernel_2))
    print('strides_2: {}'.format(strides_2))
    print('pool_2: {}'.format(pool_2))
    print('kernel_3: {}'.format(kernel_3))
    print('strides_3: {}'.format(strides_3))
    print('pool_3: {}'.format(pool_3))
    print('kernel_4: {}'.format(kernel_4))
    print('strides_4: {}'.format(strides_4))
    print('pool_4: {}'.format(pool_4))
    print('z_dimension: {}'.format(z_dimension))
    print('n_modes: {}'.format(n_modes))
    print('n_filters_1: {}'.format(params['n_filters_r1'][0]))
    print('n_filters_2: {}'.format(params['n_filters_r1'][1]))
    print('n_filters_3: {}'.format(params['n_filters_r1'][2]))
    print('n_filters_4: {}'.format(params['n_filters_r1'][3]))
    print('batch_size: {}'.format(batch_size))
    print('n_weights_r1_1: {}'.format(n_weights_fc_1))
    print('n_weights_r1_2: {}'.format(n_weights_fc_2))
    print('n_weights_r1_3: {}'.format(n_weights_fc_3))
    print()

    start_time = time.time()
    print('start time: {}'.format(strftime('%X %x %Z'))) 
    # Train model with given hyperparameters
    VICI_loss, VICI_session, VICI_saver, VICI_savedir = VICI_inverse_model.train(params, x_data_train, y_data_train,
                             x_data_test, y_data_test, y_data_test_noisefree,
                             y_normscale,
                             "inverse_model_dir_%s/inverse_model.ckpt" % params['run_label'],
                             x_data_test, bounds, fixed_vals,
                             XS_all)

    end_time = time.time()
    print('Run time : {} h'.format((end_time-start_time)/3600))

    # Print the loss.
    print()
    print("Total loss: {0:.2}".format(VICI_loss))
    print()

    # update variable outside of this function using global keyword
    global best_loss

    # save model if new best model
    if VICI_loss < best_loss:

        # Save model 
        save_path = VICI_saver.save(VICI_session,VICI_savedir)

        # save hyperparameters
        converged_hyperpar_dict = dict(filter_size = params['filter_size_r1'],
                                       conv_strides = params['conv_strides_r1'],
                                       maxpool = params['maxpool_r1'],
                                       pool_strides = params['pool_strides_r1'],
                                       z_dimension = params['z_dimension'],
                                       n_modes = params['n_modes'],
                                       n_filters = params['n_filters_r1'],
                                       batch_size = params['batch_size'],
                                       n_weights_fc = params['n_weights_r1'],
                                       best_loss = best_loss)
                                       #n_weights_r2 = params['n_weights_r2'],
                                       #n_weights_q = params['n_weights_q'])

        f = open("inverse_model_dir_%s/converged_hyperparams.txt" % params['run_label'],"w")
        f.write( str(converged_hyperpar_dict) )
        f.close()

        # update the best loss
        best_loss = VICI_loss
        
        # Print the loss.
        print()
        print("New best loss: {0:.2}".format(best_loss))
        print()

    # clear tensorflow session
    VICI_session.close()

    return VICI_loss

def gen_train(params=params,bounds=bounds,fixed_vals=fixed_vals):
    """ Generate training samples

    Parameters
    ----------
    params: dict
        Dictionary containing run parameters
    bounds: dict
        Dictionary containing allowed bounds of GW source parameters
    fixed_vals: dict
        Dictionary containing the fixed values of GW source parameters
    """

    # Check for requried parameters files
    if params == None or bounds == None or fixed_vals == None:
        print('Missing either params file, bounds file or fixed vals file')
        exit()

    # Load parameters files
    with open(params, 'r') as fp:
        params = json.load(fp)
    with open(bounds, 'r') as fp:
        bounds = json.load(fp)
    with open(fixed_vals, 'r') as fp:
        fixed_vals = json.load(fp)

    # Make training set directory
    os.system('mkdir -p %s' % params['train_set_dir'])

    # Make directory for plots
    os.system('mkdir -p %s/latest_%s' % (params['plot_dir'],params['run_label']))

    print()
    print('... Making training set')
    print()

    # Iterate over number of requested training samples
    for i in range(0,params['tot_dataset_size'],params['tset_split']):

        logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': True,
        })
        with suppress_stdout():
            # generate training sample source parameter, waveform and snr
            signal_train, signal_train_pars,snrs = run(sampling_frequency=params['ndata']/params['duration'],
                                                          duration=params['duration'],
                                                          N_gen=params['tset_split'],
                                                          ref_geocent_time=params['ref_geocent_time'],
                                                          bounds=bounds,
                                                          fixed_vals=fixed_vals,
                                                          rand_pars=params['rand_pars'],
                                                          seed=params['training_data_seed']+i,
                                                          label=params['run_label'],
                                                          training=True,det=params['det'],
                                                          psd_files=params['psd_files'])
        logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        })
        print("Generated: %s/data_%d-%d.h5py ..." % (params['train_set_dir'],(i+params['tset_split']),params['tot_dataset_size']))

        # store training sample information in hdf5 format
        hf = h5py.File('%s/data_%d-%d.h5py' % (params['train_set_dir'],(i+params['tset_split']),params['tot_dataset_size']), 'w')
        for k, v in params.items():
            try:
                hf.create_dataset(k,data=v)
            except:
                pass
        hf.create_dataset('x_data', data=signal_train_pars)
        for k, v in bounds.items():
            hf.create_dataset(k,data=v)
        hf.create_dataset('y_data_noisy', data=np.array([]))
        hf.create_dataset('y_data_noisefree', data=signal_train)
        hf.create_dataset('rand_pars', data=np.string_(params['rand_pars']))
        hf.create_dataset('snrs', data=snrs)
        hf.close()
    return

def gen_test(params=params,bounds=bounds,fixed_vals=fixed_vals):
    """ Generate testing sample time series and posteriors using Bayesian inference (bilby)

    Parameters
    ----------
    params: dict
        Dictionary containing run parameters
    bounds: dict
        Dictionary containing allowed bounds of GW source parameters
    fixed_vals: dict
        Dictionary containing the fixed values of GW source parameters
    """

    # Check for requried parameters files
    if params == None or bounds == None or fixed_vals == None:
        print('Missing either params file, bounds file or fixed vals file')
        exit()

    # Load parameters files
    with open(params, 'r') as fp:
        params = json.load(fp)
    with open(bounds, 'r') as fp:
        bounds = json.load(fp)
    with open(fixed_vals, 'r') as fp:
        fixed_vals = json.load(fp)

    # Make testing set directory
    os.system('mkdir -p %s' % params['test_set_dir'])

    # Make testing samples
    for i in range(params['r']):
        temp_noisy, temp_noisefree, temp_pars, temp_snr = run(sampling_frequency=params['ndata']/params['duration'],
                                                      duration=params['duration'],
                                                      N_gen=1,
                                                      ref_geocent_time=params['ref_geocent_time'],
                                                      bounds=bounds,
                                                      fixed_vals=fixed_vals,
                                                      rand_pars=params['rand_pars'],
                                                      inf_pars=params['inf_pars'],
                                                      label=params['bilby_results_label'] + '_' + str(i),
                                                      out_dir=params['pe_dir'],
                                                      samplers=params['samplers'],
                                                      training=False,
                                                      seed=params['testing_data_seed']+i,
                                                      do_pe=params['doPE'],det=params['det'],
                                                      psd_files=params['psd_files'])

        signal_test_noisy = temp_noisy
        signal_test_noisefree = temp_noisefree
        signal_test_pars = temp_pars
        signal_test_snr = temp_snr

        print("Generated: %s/data_%s.h5py ..." % (params['test_set_dir'],params['run_label']))

        # Save generated testing samples in h5py format
        hf = h5py.File('%s/data_%d.h5py' % (params['test_set_dir'],i),'w')
        for k, v in params.items():
            try:
                hf.create_dataset(k,data=v)
            except:
                pass
        hf.create_dataset('x_data', data=signal_test_pars)
        for k, v in bounds.items():
            hf.create_dataset(k,data=v)
        hf.create_dataset('y_data_noisefree', data=signal_test_noisefree)
        hf.create_dataset('y_data_noisy', data=signal_test_noisy)
        hf.create_dataset('rand_pars', data=np.string_(params['rand_pars']))
        hf.create_dataset('snrs', data=signal_test_snr)
        hf.close()
    return

def train(params=params,bounds=bounds,fixed_vals=fixed_vals,resume_training=False):
    """ Train neural network given pre-made training/testing samples

    Parameters
    ----------
    params: dict
        Dictionary containing run parameters
    bounds: dict
        Dictionary containing allowed bounds of GW source parameters
    fixed_vals: dict
        Dictionary containing the fixed values of GW source parameters
    resume_training: bool
        If True, continue training a pre-trained model.
    """

    # Check for requried parameters files
    if params == None or bounds == None or fixed_vals == None:
        print('Missing either params file, bounds file or fixed vals file')
        exit()

    # Load parameters files
    with open(params, 'r') as fp:
        params = json.load(fp)
    with open(bounds, 'r') as fp:
        bounds = json.load(fp)
    with open(fixed_vals, 'r') as fp:
        fixed_vals = json.load(fp)

    # define which gpu to use during training
    gpu_num = str(params['gpu_num'])                                            # first GPU used by default
    os.environ["CUDA_VISIBLE_DEVICES"]=gpu_num

    # Let GPU consumption grow as needed
    config = tf.compat.v1.ConfigProto()
    config.gpu_options.allow_growth = True
    session = tf.compat.v1.Session(config=config)

    # If resuming training, set KL ramp off
    if resume_training:
        params['resume_training'] = True
        params['ramp'] = False

    # load the noisefree training data back in
    x_data_train, y_data_train, _, y_normscale, snrs_train = load_data(params,bounds,fixed_vals,params['train_set_dir'],params['inf_pars'])

    # load the noisy testing data back in
    x_data_test, y_data_test_noisefree, y_data_test,_,snrs_test = load_data(params,bounds,fixed_vals,params['test_set_dir'],params['inf_pars'],load_condor=True)

    # reshape time series arrays for single channel ( N_samples,fs*duration,n_detectors -> (N_samples,fs*duration*n_detectors) )
    y_data_train = y_data_train.reshape(y_data_train.shape[0]*y_data_train.shape[1],y_data_train.shape[2]*y_data_train.shape[3])
#    y_data_train = y_data_train.reshape(y_data_train.shape[1],y_data_train.shape[2]*y_data_train.shape[3])
    y_data_test = y_data_test.reshape(y_data_test.shape[0],y_data_test.shape[1]*y_data_test.shape[2])
    y_data_test_noisefree = y_data_test_noisefree.reshape(y_data_test_noisefree.shape[0],y_data_test_noisefree.shape[1]*y_data_test_noisefree.shape[2])

    # Make directory for plots
    os.system('mkdir -p %s/latest_%s' % (params['plot_dir'],params['run_label']))

    # Save configuration file to public_html directory
    f = open('%s/latest_%s/params_%s.txt' % (params['plot_dir'],params['run_label'],params['run_label']),"w")
    f.write( str(params) )
    f.close()

    # load up the posterior samples (if they exist)
    if params['pe_dir'] != None:
        # load generated samples back in
        post_files = []
        #~/bilby_outputs/bilby_output_dynesty1/multi-modal3_0.h5py

        # first identify directory with lowest number of total finished posteriors
        num_finished_post = int(1e8)
        for i in params['samplers']:
            if i == 'vitamin':
                continue

            # remove any remaining resume files
            resume_files=glob.glob('%s_%s1/*.resume*' % (params['pe_dir'],i))
            filelist = [resume_files]
            for file_idx,file_type in enumerate(filelist):
                for file in file_type:
                    os.remove(file) 

            for j in range(1):
                input_dir = '%s_%s%d/' % (params['pe_dir'],i,j+1)
                if type("%s" % input_dir) is str:
                    dataLocations = ["%s" % input_dir]

                filenames = sorted(os.listdir(dataLocations[0]), key=lambda x: int(x.split('.')[0].split('_')[-1]))      
                if len(filenames) < num_finished_post:
                    sampler_loc = i + str(j+1)
                    num_finished_post = len(filenames)


        dataLocations_try = '%s_%s' % (params['pe_dir'],sampler_loc)
        dataLocations = '%s_%s1' % (params['pe_dir'],params['samplers'][1])

        i_idx = 0
        i = 0
        i_idx_use = []

        # Iterate over requested number of testing samples to use
        while i_idx < params['r']:

            filename_try = '%s/%s_%d.h5py' % (dataLocations_try,params['bilby_results_label'],i)
            filename = '%s/%s_%d.h5py' % (dataLocations,params['bilby_results_label'],i)

            # Assert user has the minimum number of test samples generated
            number_of_files_in_dir = len(os.listdir(dataLocations))
            try:
                assert number_of_files_in_dir >= params['r']
            except Exception as e:
                print(e)
                print('You are requesting to use more test GW time series than you have made.')
                print('... Exiting program now')
                exit()

            # If file does not exist, skip to next file
            try:
                h5py.File(filename_try, 'r')
            except Exception as e:
                i+=1
                print(e)
                continue

            print()
            print('... Loading test sample -> ' + filename)
            post_files.append(filename)
            data_temp = {} 
            n = 0
       
            # Retrieve all source parameters to do inference on
            for q in params['inf_pars']:
                 p = q + '_post'
                 par_min = q + '_min'
                 par_max = q + '_max'
                 data_temp[p] = h5py.File(filename, 'r')[p][:]
                 if p == 'psi_post':
                     data_temp[p] = np.remainder(data_temp[p],np.pi)
                 if p == 'geocent_time_post':
                     data_temp[p] = data_temp[p] - params['ref_geocent_time']
                 data_temp[p] = (data_temp[p] - bounds[par_min]) / (bounds[par_max] - bounds[par_min])
                 Nsamp = data_temp[p].shape[0]
                 n = n + 1
            XS = np.zeros((Nsamp,n))
            j = 0

            # place retrieved source parameters in numpy array rather than dictionary
            for p,d in data_temp.items():
                XS[:,j] = d
                j += 1

            # Append test sample posteriors to existing array of other test sample posteriors
            if i_idx == 0:
                XS_all = np.expand_dims(XS[:params['n_samples'],:], axis=0)
            else:
                XS_all = np.vstack((XS_all,np.expand_dims(XS[:params['n_samples'],:], axis=0)))


            # add index to mark progress through while loop
            i_idx_use.append(i_idx)
            i+=1
            i_idx+=1


        # Identify test samples that are present accross all Bayesian PE samplers
        y_data_test = y_data_test[i_idx_use,:]
        y_data_test_noisefree = y_data_test_noisefree[i_idx_use,:]
        x_data_test = x_data_test[i_idx_use,:]
    # Set posterior samples to None if posteriors don't exist
    elif params['pe_dir'] == None:
        XS_all = None

    # reshape y data into channels last format for convolutional approach (if requested)
    if params['n_filters_r1'] != None:
        y_data_test_copy = np.zeros((y_data_test.shape[0],params['ndata'],len(params['det'])))
        y_data_test_noisefree_copy = np.zeros((y_data_test_noisefree.shape[0],params['ndata'],len(params['det'])))
        y_data_train_copy = np.zeros((y_data_train.shape[0],params['ndata'],len(params['det'])))
        for i in range(y_data_test.shape[0]):
            for j in range(len(params['det'])):
                idx_range = np.linspace(int(j*params['ndata']),int((j+1)*params['ndata'])-1,num=params['ndata'],dtype=int)
                y_data_test_copy[i,:,j] = y_data_test[i,idx_range]
                y_data_test_noisefree_copy[i,:,j] = y_data_test_noisefree[i,idx_range]
        y_data_test = y_data_test_copy
        y_data_noisefree_test = y_data_test_noisefree_copy

        for i in range(y_data_train.shape[0]):
            for j in range(len(params['det'])):
                idx_range = np.linspace(int(j*params['ndata']),int((j+1)*params['ndata'])-1,num=params['ndata'],dtype=int)
                y_data_train_copy[i,:,j] = y_data_train[i,idx_range]
        y_data_train = y_data_train_copy

    # run hyperparameter optimization
    if params['hyperparam_optim'] == True:

        # Run optimization
        search_result = gp_minimize(func=hyperparam_fitness,
                            dimensions=dimensions,
                            acq_func='EI', # Negative Expected Improvement.
                            n_calls=params['hyperparam_n_call'],
                            x0=default_hyperparams)

        from skopt import dump
        dump(search_result, 'search_result_store')

        # plot best loss as a function of optimization step
        plt.close('all')
        plot_convergence(search_result)
        plt.savefig('%s/latest_%s/hyperpar_convergence.png' % (params['plot_dir'],params['run_label']))
        print('... Saved hyperparameter convergence loss to -> %s/latest_%s/hyperpar_convergence.png' % (params['plot_dir'],params['run_label']))
        print('... Did a hyperparameter search') 

    # train using user defined params
    else:
        CVAE_model.train(params, x_data_train, y_data_train,
                                 x_data_test, y_data_test, y_data_test_noisefree,
                                 y_normscale,
                                 "inverse_model_dir_%s/inverse_model.ckpt" % params['run_label'],
                                 x_data_test, bounds, fixed_vals,
                                 XS_all,snrs_test) 
    return

# if we are now testing the network
def test(params=params,bounds=bounds,fixed_vals=fixed_vals,use_gpu=False):
    """ Test a pre-trained neural network. There are several metrics by 
    which the user may test the efficiency of the model (e.g. KL divergence, 
    pp plots, corner plots).

    Parameters
    ----------
    params: dict
        Dictionary containing run parameters
    bounds: dict
        Dictionary containing allowed bounds of GW source parameters
    fixed_vals: dict
        Dictionary containing the fixed values of GW source parameters
    use_gpu: bool
        If True, use a GPU to generate samples from the posterior of a pretrained neural network
    """

    # Check for requried parameters files
    if params == None or bounds == None or fixed_vals == None:
        print('Missing either params file, bounds file or fixed vals file')
        exit()

    # Load parameters files
    with open(params, 'r') as fp:
        params = json.load(fp)
    with open(bounds, 'r') as fp:
        bounds = json.load(fp)
    with open(fixed_vals, 'r') as fp:
        fixed_vals = json.load(fp)

    if use_gpu == True:
        print("... GPU found")
        os.environ["CUDA_VISIBLE_DEVICES"]=str(params['gpu_num'])
        config = tf.compat.v1.ConfigProto()
        config.gpu_options.allow_growth = True  # Let GPU consumption grow as needed
        session = tf.compat.v1.Session(config=config)
    else:
        print("... Using CPU")
        os.environ["CUDA_VISIBLE_DEVICES"]=''
        config = tf.compat.v1.ConfigProto()
        session = tf.compat.v1.Session(config=config)

    y_normscale = params['y_normscale']

    # load the testing data time series and source parameter truths
    x_data_test, y_data_test_noisefree, y_data_test,_,snrs_test = load_data(params,bounds,fixed_vals,params['test_set_dir'],params['inf_pars'],load_condor=True)

    # Make directory to store plots
    os.system('mkdir -p %s/latest_%s' % (params['plot_dir'],params['run_label']))

    # reshape arrays for single channel network (this will be overwritten if channels last is requested by user)
    y_data_test = y_data_test.reshape(y_data_test.shape[0],y_data_test.shape[1]*y_data_test.shape[2])
    y_data_test_noisefree = y_data_test_noisefree.reshape(y_data_test_noisefree.shape[0],y_data_test_noisefree.shape[1]*y_data_test_noisefree.shape[2])

    # Make directory for plots
    os.system('mkdir -p %s/latest_%s' % (params['plot_dir'],params['run_label']))

    # load up the posterior samples (if they exist)
    # load generated samples back in
    post_files = []
    #~/bilby_outputs/bilby_output_dynesty1/multi-modal3_0.h5py

    # Identify directory with lowest number of total finished posteriors
    num_finished_post = int(1e8)
    for i in params['samplers']:
        if i == 'vitamin':# or i == 'emcee':
            continue

        # remove any remaining resume files
        resume_files=glob.glob('%s_%s1/*.resume*' % (params['pe_dir'],i))
        filelist = [resume_files]
        for file_idx,file_type in enumerate(filelist):
            for file in file_type:
                os.remove(file)

        for j in range(1):
            input_dir = '%s_%s%d/' % (params['pe_dir'],i,j+1)
            if type("%s" % input_dir) is str:
                dataLocations = ["%s" % input_dir]

            filenames = sorted(os.listdir(dataLocations[0]), key=lambda x: int(x.split('.')[0].split('_')[-1]))
            if len(filenames) < num_finished_post:
                sampler_loc = i + str(j+1)
                num_finished_post = len(filenames)


    # Assert user has the minimum number of test samples generated
    number_of_files_in_dir = len(os.listdir(dataLocations[0]))
    try:
        assert number_of_files_in_dir >= params['r']
    except Exception as e:
        print(e)
        print('You are requesting to use more GW time series than you have made.')
        exit()

    samp_posteriors = {}
    # Iterate over all Bayesian PE samplers
    for samp_idx in params['samplers'][1:]:
        dataLocations_try = '%s_%s' % (params['pe_dir'],sampler_loc)
        dataLocations = '%s_%s' % (params['pe_dir'],samp_idx+'1')
        i_idx = 0
        i = 0
        i_idx_use = []
        x_data_test_unnorm = np.copy(x_data_test)


        # Iterate over all requested testing samples
        while i_idx < params['r']:


            filename_try = '%s/%s_%d.h5py' % (dataLocations_try,params['bilby_results_label'],i)
            filename = '%s/%s_%d.h5py' % (dataLocations,params['bilby_results_label'],i)

            # If file does not exist, skip to next file
            try:
                h5py.File(filename_try, 'r')
            except Exception as e:
                i+=1
                print(e)
                continue

            print('... Loading test sample file -> ' + filename)
            post_files.append(filename)
            
            # Prune emcee samples for bad likelihood chains
            if samp_idx == 'emcee':
                emcee_pruned_samples = prune_samples(filename,params)

            data_temp = {}
            n = 0
            for q_idx,q in enumerate(params['inf_pars']):
                 p = q + '_post'
                 par_min = q + '_min'
                 par_max = q + '_max'
                 if p == 'psi_post':
                     data_temp[p] = np.remainder(data_temp[p],np.pi)

                 if samp_idx == 'emcee':
                     data_temp[p] = emcee_pruned_samples[:,q_idx]
                 else:
                     data_temp[p] = np.float64(h5py.File(filename, 'r')[p][:])

                 if p == 'geocent_time_post' or p == 'geocent_time_post_with_cut':
                     data_temp[p] = np.subtract(np.float64(data_temp[p]),np.float64(params['ref_geocent_time'])) 

                 Nsamp = data_temp[p].shape[0]
                 n = n + 1

            XS = np.zeros((Nsamp,n))
            j = 0

            # store posteriors in numpy array rather than dictionary
            for p,d in data_temp.items():
                XS[:,j] = d
                j += 1

            rand_idx_posterior = np.random.choice(np.linspace(0,XS.shape[0]-1,dtype=np.int),params['n_samples'])
            # Append test sample posterior to existing array of test sample posteriors
            if i_idx == 0:
                #XS_all = np.expand_dims(XS[rand_idx_posterior,:], axis=0)
                XS_all = np.expand_dims(XS[:params['n_samples'],:], axis=0)
            else:
                try:
                    #XS_all = np.vstack((XS_all,np.expand_dims(XS[rand_idx_posterior,:], axis=0)))
                    XS_all = np.vstack((XS_all,np.expand_dims(XS[:params['n_samples'],:], axis=0)))
                except ValueError as error: # If not enough posterior samples, exit with ValueError
                    print('Not enough samples from the posterior generated')
                    print(error)
                    exit()

            # Get unnormalized array with source parameter truths
            for q_idx,q in enumerate(params['inf_pars']):
                par_min = q + '_min'
                par_max = q + '_max'

                x_data_test_unnorm[i_idx,q_idx] = (x_data_test_unnorm[i_idx,q_idx] * (bounds[par_max] - bounds[par_min])) + bounds[par_min]

            # Add to index in order to progress through while loop iterating over testing samples
            i_idx_use.append(i_idx)
            i+=1
            i_idx+=1

        # Add all testing samples for current Bayesian PE sampler to dictionary of all other Bayesian PE sampler test samples
        samp_posteriors[samp_idx+'1'] = XS_all

    # Ensure no failed test sample Bayesian PE runs are used
    x_data_test = x_data_test[i_idx_use,:]
    x_data_test_unnorm = x_data_test_unnorm[i_idx_use,:]
    y_data_test = y_data_test[i_idx_use,:]
    y_data_test_noisefree = y_data_test_noisefree[i_idx_use,:]

    # reshape y data into channels last format for convolutional approach
    y_data_test_copy = np.zeros((y_data_test.shape[0],params['ndata'],len(params['det'])))
    if params['n_filters_r1'] != None:
        for i in range(y_data_test.shape[0]):
            for j in range(len(params['det'])):
                idx_range = np.linspace(int(j*params['ndata']),int((j+1)*params['ndata'])-1,num=params['ndata'],dtype=int)
                y_data_test_copy[i,:,j] = y_data_test[i,idx_range]
        y_data_test = y_data_test_copy
    

    # Reshape time series  array to right format for 1-channel configuration
    if params['by_channel'] == False:
        y_data_test_new = []
        for sig in y_data_test:
            y_data_test_new.append(sig.T)
        y_data_test = np.array(y_data_test_new)
        del y_data_test_new

    # check is basemap is installed
    if not skyplotting_usage:
        params['Make_sky_plot'] = False

    VI_pred_all = []
    # Iterate over total number of testing samples
    for i in range(params['r']):

        # If True, continue through and make corner plots
        if params['make_corner_plots'] == False:
            break
        
        # Generate ML posteriors using pre-trained model
        if params['n_filters_r1'] != None: # for convolutional approach
             VI_pred, dt, _  = CVAE_model.run(params, np.expand_dims(y_data_test[i],axis=0), np.shape(x_data_test)[1],
                                                         y_normscale,
                                                         "inverse_model_dir_%s/inverse_model.ckpt" % params['run_label'])
        else:                                                          # for fully-connected approach
            VI_pred, dt, _  = CVAE_model.run(params, y_data_test[i].reshape([1,-1]), np.shape(x_data_test)[1],
                                                         y_normscale,
                                                         "inverse_model_dir_%s/inverse_model.ckpt" % params['run_label'])

        # Make corner corner plots
        bins=50
       
        # Define default corner plot arguments
        defaults_kwargs = dict(
                    bins=bins, smooth=0.9, label_kwargs=dict(fontsize=16),
                    title_kwargs=dict(fontsize=16), show_titles=False,
                    truth_color='black', quantiles=None,#[0.16, 0.84],
                    levels=(0.50,0.90), density=True, stacked=True,
                    plot_density=False, plot_datapoints=True,
                    max_n_ticks=3)

        matplotlib.rc('text', usetex=True)                
        parnames = []
    
        # Get infered parameter latex labels for corner plot
        for k_idx,k in enumerate(params['rand_pars']):
            if np.isin(k, params['inf_pars']):
                parnames.append(params['corner_labels'][k])

        # unnormalize the predictions from VICI (comment out if not wanted)
        color_cycle=['tab:blue','tab:green','tab:purple','tab:orange']
        legend_color_cycle=['blue','green','purple','orange']
        for q_idx,q in enumerate(params['inf_pars']):
                par_min = q + '_min'
                par_max = q + '_max'
                VI_pred[:,q_idx] = (VI_pred[:,q_idx] * (bounds[par_max] - bounds[par_min])) + bounds[par_min]

        # Convert hour angle to right ascension
        if np.isin('ra', params['inf_pars']) and  np.isin('geocent_time', params['inf_pars']):
            for k_idx, k in enumerate(params['inf_pars']):
                if k == 'geocent_time':
                    geo_idx = k_idx
                elif k=='ra':
                    ra_idx=k_idx
            # Check if both geocentime and RA exist
            try:
                geo_idx; ra_idx
            except NameError:
                print('Either time or RA is fixed. Not converting RA to hour angle.')
            else:
                for k_idx in range(VI_pred.shape[0]):
                    VI_pred[k_idx,ra_idx]=np.mod(GreenwichMeanSiderealTime(float(params['ref_geocent_time']+VI_pred[k_idx,geo_idx])) - VI_pred[k_idx,ra_idx], 2.0*np.pi)


        # Iterate over all Bayesian PE samplers and plot results
        custom_lines = []
        truths = x_data_test_unnorm[i,:]
        for samp_idx,samp in enumerate(params['samplers'][1:]):

            bilby_pred = samp_posteriors[samp+'1'][i]

            # compute weights, otherwise the 1d histograms will be different scales, could remove this
            weights = np.ones(len(VI_pred)) * (len(samp_posteriors[samp+'1'][i]) / len(VI_pred))
            if samp_idx == 0:
                figure = corner.corner(bilby_pred,**defaults_kwargs,labels=parnames,
                               color=color_cycle[samp_idx],
                               truths=truths
                               )
            else:
                figure = corner.corner(bilby_pred,**defaults_kwargs,labels=parnames,
                               color=color_cycle[samp_idx],
                               truths=truths,
                               fig=figure)
            custom_lines.append(Line2D([0], [0], color=legend_color_cycle[samp_idx], lw=4))

        # plot predicted ML results
        corner.corner(VI_pred, **defaults_kwargs, labels=parnames,
                           color='tab:red', fill_contours=True,
                           fig=figure)
        custom_lines.append(Line2D([0], [0], color='red', lw=4))

        if params['Make_sky_plot'] == True:
            # Compute skyplot
#            left, bottom, width, height = [0.55, 0.47, 0.5, 0.39] # orthographic representation
#            left, bottom, width, height = [0.525, 0.47, 0.45, 0.44] # mollweide representation
            print()
            print('... Generating sky plot')
            print()
            left, bottom, width, height = [0.46, 0.6, 0.5, 0.5] # switch with waveform positioning
            ax_sky = figure.add_axes([left, bottom, width, height]) 

            sky_color_cycle=['blue','green','purple','orange']
            sky_color_map_cycle=['Blues','Greens','Purples','Oranges']
            for samp_idx,samp in enumerate(params['samplers'][1:]):
                bilby_pred = samp_posteriors[samp+'1'][i]
                if samp_idx == 0:
                    ax_sky = plot_sky(bilby_pred[:,-2:],filled=False,cmap=sky_color_map_cycle[samp_idx],col=sky_color_cycle[samp_idx],trueloc=truths[-2:])
                else:
                    ax_sky = plot_sky(bilby_pred[:,-2:],filled=False,cmap=sky_color_map_cycle[samp_idx],col=sky_color_cycle[samp_idx], trueloc=truths[-2:], ax=ax_sky)
            ax_sky = plot_sky(VI_pred[:,-2:],filled=True,trueloc=truths[-2:],ax=ax_sky)


        #left, bottom, width, height = [0.34, 0.82, 0.3, 0.17] # standard positioning
        left, bottom, width, height = [0.67, 0.48, 0.3, 0.2] # swtiched with skymap positioning
        ax2 = figure.add_axes([left, bottom, width, height])
        # plot waveform in upper-right hand corner
        ax2.plot(np.linspace(0,1,params['ndata']),y_data_test_noisefree[i,:params['ndata']],color='cyan',zorder=50)
        snr = round(snrs_test[i,0],2)
        if params['n_filters_r1'] != None:
            if params['by_channel'] == False:
                 ax2.plot(np.linspace(0,1,params['ndata']),y_data_test[i,0,:params['ndata']],color='darkblue')#,label='SNR: '+str(snr))
            else:
                ax2.plot(np.linspace(0,1,params['ndata']),y_data_test[i,:params['ndata'],0],color='darkblue')#,label='SNR: '+str(snr))
        else:
            ax2.plot(np.linspace(0,1,params['ndata']),y_data_test[i,:params['ndata']],color='darkblue')#,label='SNR: '+str(snr))
        ax2.set_xlabel(r"$\textrm{time (seconds)}$",fontsize=16)
        ax2.yaxis.set_visible(False)
        ax2.tick_params(axis="x", labelsize=12)
        ax2.tick_params(axis="y", labelsize=12)
        ax2.set_ylim([-6,6])
        ax2.grid(False)
        ax2.margins(x=0,y=0)

        # Save corner plot to latest public_html directory
        figure.legend(handles=custom_lines, labels=['Dynesty', 'ptemcee', 'VItamin'],
                      loc=(0.86,0.22), fontsize=20)
        plt.savefig('%s/latest_%s/corner_plot_%s_%d.png' % (params['plot_dir'],params['run_label'],params['run_label'],i))
        plt.close()
        del figure
        print()
        print('... Made corner plot: %s' % str(i+1))
        print('... Saved corner plot to -> %s/latest_%s/corner_plot_%s_%d.png' % (params['plot_dir'],params['run_label'],params['run_label'],i))
        print()

        # Store ML predictions for later plotting use
        VI_pred_all.append(VI_pred)

    VI_pred_all = np.array(VI_pred_all)

    # Define pp and KL plotting class
#    XS_all = None; x_data_test = None; y_data_test = None; y_normscale = None; snrs_test = None
    plotter = plotting.make_plots(params,XS_all,VI_pred_all,x_data_test)

    if params['make_kl_plot'] == True:    
        # Make KL plots
        plotter.gen_kl_plots(CVAE_model,y_data_test,x_data_test,y_normscale,bounds,snrs_test)

    # Make bilby pp plot
#    plotter.plot_bilby_pp(VICI_inverse_model,y_data_test,x_data_test,0,y_normscale,x_data_test,bounds)
#    exit()

    if params['make_pp_plot'] == True:
        # Make pp plot
        plotter.plot_pp(CVAE_model,y_data_test,x_data_test,y_normscale,bounds)

    if params['make_loss_plot'] == True:
        plotter.plot_loss()

    return

def gen_samples(params=params,bounds=bounds,fixed_vals=fixed_vals,model_loc='model_ex/model.ckpt',test_set='test_waveforms/',num_samples=None,plot_corner=True,use_gpu=False):
    """ Function to generate VItamin samples given a trained model

    Parameters
    ----------
    params: dict
        Dictionary containing run parameters
    bounds: dict
        Dictionary containing allowed bounds of GW source parameters
    fixed_vals: dict
        Dictionary containing the fixed values of GW source parameters
    model_loc: str
        location of pre-trained model (i.e. file with .ckpt)
    test_set: str
        dictionary location of test sample time series
    num_samples: float
        number of posterior samples to generate using neural network
    plot_corner: bool
        if true, make corner plots of generated posterior samples
    use_gpu: bool
        if true, use gpu to make posterior samples

    Returns
    -------
    samples: array_like
        posterior samples generated by neural network
    """

    # Check for requried parameters files
    if params == None or bounds == None or fixed_vals == None:
        print('Missing either params file, bounds file or fixed vals file')
        exit()

    # Load parameters files
    with open(params, 'r') as fp:
        params = json.load(fp)
    with open(bounds, 'r') as fp:
        bounds = json.load(fp)
    with open(fixed_vals, 'r') as fp:
        fixed_vals = json.load(fp)

    if use_gpu == True:
        print('GPU found')
        os.environ["CUDA_VISIBLE_DEVICES"]=str(params['gpu_num'])
        config = tf.compat.v1.ConfigProto()
        config.gpu_options.allow_growth = True  # Let GPU consumption grow as needed
    else:
        print("No GPU found")
        os.environ["CUDA_VISIBLE_DEVICES"]=''
        config = tf.compat.v1.ConfigProto() 

    session = tf.compat.v1.Session(config=config)

    if num_samples != None:
        params['n_samples'] = num_samples

    # load generated samples
    files = []
   
    # Get list of all training/testing files and define dictionary to store values in files
    if type("%s" % test_set) is str:
        dataLocations = ["%s" % test_set]
        data={'y_data_noisy': []}

    # Sort files from first generated to last generated
    filenames = sorted(os.listdir(dataLocations[0]), key=lambda x: int(x.split('.')[0].split('_')[-1]))

    # Append training/testing filenames to list. Ignore those that can't be loaded
    for filename in filenames:
        try:
            files.append(filename)
        except OSError:
            print('Could not load requested file')
            continue

    # Iterate over all training/testing files and store source parameters, time series and SNR info in dictionary
    for filename in files:
        try:
            data_temp={'y_data_noisy': h5py.File(dataLocations[0]+'/'+filename, 'r')['y_data_noisy'][:]}
            data['y_data_noisy'].append(np.expand_dims(data_temp['y_data_noisy'], axis=0))
        except OSError:
            print('Could not load requested file')
            continue

    # Extract the prior bounds from training/testing files
    data['y_data_noisy'] = np.concatenate(np.array(data['y_data_noisy']), axis=0)
    

    y_data_test = data['y_data_noisy']

    # Define time series normalization factor to use on test samples. We consistantly use the same normscale value if loading by chunks
    y_normscale = params['y_normscale']   
 
    y_data_test = y_data_test.reshape(y_data_test.shape[0],y_data_test.shape[1]*y_data_test.shape[2])
    # reshape y data into channels last format for convolutional approach
    y_data_test_copy = np.zeros((y_data_test.shape[0],params['ndata'],len(params['det'])))
    if params['n_filters_r1'] != None:
        for i in range(y_data_test.shape[0]):
            for j in range(len(params['det'])):
                idx_range = np.linspace(int(j*params['ndata']),int((j+1)*params['ndata'])-1,num=params['ndata'],dtype=int)       
                y_data_test_copy[i,:,j] = y_data_test[i,idx_range]
        y_data_test = y_data_test_copy
    num_timeseries=y_data_test.shape[0]
    

    """
    # Daniel Williams waveforms
    num_timeseries = 1
    f = open('daniel_williams_BNU_waveforms/hunter-mdc/0.json')
    data = json.load(f)
    y_data_test = np.zeros((1,params['ndata'],len(params['det'])))
    for det_idx,det in enumerate(params['det']):
        y_data_test[0,-(np.array(data['data'][det]).shape[0]):,det_idx] = np.array(data['data'][det])

    # Get x_data_test
    x_data_test = []
    for idx, i in enumerate(params['inf_pars']):
        if i == data['meta'][i]:
            x_data_test.append(data['meta'][i])
    x_data_test = np.array(x_data_test)
    print(x_data_test)
    exit()

    # whiten Daniel waveforms
    # Set up interferometers. These default to their design
    # sensitivity

    # define the start time of the timeseries
    start_time = params['ref_geocent_time']-params['duration']/2.0

    ifos = bilby.gw.detector.InterferometerList(params['det'])
    # set noise to be colored Gaussian noise
    ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=params['ndata'], duration=params['duration'],
    start_time=start_time)

    for det_idx,det in enumerate(params['det']):
        fft_waveform = np.fft.rfft(y_data_test[0,:,det_idx])
        whitened_signal = fft_waveform / ifos[det_idx].amplitude_spectral_density_array
        y_data_test[0,:,det_idx] = np.fft.irfft(whitened_signal) + np.random.normal(loc=0,scale=1.0,size=(params['ndata']))
    """

    samples = np.zeros((num_timeseries,num_samples,len(params['inf_pars'])))
    for i in range(num_timeseries):
        samples[i,:], dt, _  = CVAE_model.run(params, np.expand_dims(y_data_test[i],axis=0), len(params['inf_pars']),
                                                              params['y_normscale'],
                                                              model_loc)
        print('... Runtime to generate samples is: ' + str(dt))

        # convert RA to hour angle for test set validation cost if both ra and geo time present
        if np.isin('ra', params['inf_pars']) and  np.isin('geocent_time', params['inf_pars']):     
            # get geocenttime index
            for k_idx,k in enumerate(params['inf_pars']):
                if k == 'geocent_time':
                    geo_idx = k_idx
                elif k == 'ra':
                    ra_idx = k_idx

            # unnormalize and get gps time
            samples[i,:][:,ra_idx] = (samples[i,:][:,ra_idx] * (bounds['ra_max'] - bounds['ra_min'])) + bounds['ra_min']

            gps_time_arr = (samples[i,:][:,geo_idx] * (bounds['geocent_time_max'] - bounds['geocent_time_min'])) + bounds['geocent_time_min']
            # convert to RA
            # Iterate over all training samples and convert to hour angle
            for k in range(samples[i,:].shape[0]):
                samples[i,:][k,ra_idx]=np.mod(GreenwichMeanSiderealTime(float(params['ref_geocent_time']+gps_time_arr[k]))-samples[i,:][k,ra_idx], 2.0*np.pi)
            # normalize
            samples[i,:][:,ra_idx]=(samples[i,:][:,ra_idx] - bounds['ra_min']) / (bounds['ra_max'] - bounds['ra_min'])

        # unnormalize predictions
        for q_idx,q in enumerate(params['inf_pars']):
            par_min = q + '_min'
            par_max = q + '_max'
            samples[i,:,q_idx] = (samples[i,:,q_idx] * (bounds[par_max] - bounds[par_min])) + bounds[par_min]

        # plot results
        if plot_corner==True:
            # Get infered parameter latex labels for corner plot
            parnames=[]
            for k_idx,k in enumerate(params['rand_pars']):
                if np.isin(k, params['inf_pars']):
                    parnames.append(params['corner_labels'][k])
            figure = corner.corner(samples[i,:,:],labels=parnames)
            plt.savefig('./vitamin_corner_timeseries-%d.png' % i)
            plt.close()
            print('... Saved corner plot to -> ./vitamin_corner_timeseries-%d.png' % i)
            print()

    print('... All posterior samples generated for all waveforms in test sample directory!')
    return samples

# If running module from command line
if args.gen_train:
    gen_train(params,bounds,fixed_vals)
if args.gen_test:
    gen_test(params,bounds,fixed_vals)
if args.train:
    train(params,bounds,fixed_vals)
if args.test:
    test(params,bounds,fixed_vals,use_gpu=bool(args.use_gpu))
if args.gen_samples:
    gen_samples(params,bounds,fixed_vals,model_loc=args.pretrained_loc,
                test_set=args.test_set_loc,num_samples=args.num_samples,use_gpu=bool(args.use_gpu))




import collections

#import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import numpy as np
import math as m

from . import vae_utils

# based on implementation here:
# https://github.com/tensorflow/models/blob/master/autoencoder/autoencoder_models/VariationalAutoencoder.py

class VariationalAutoencoder(object):

    def __init__(self, name, n_input=256, n_output=4, n_channels=3, n_weights=2048, n_modes=2, n_hlayers=2, drate=0.2, n_filters=8, filter_size=8, maxpool=4, n_conv=2):
        
        self.n_input = n_input
        self.n_output = n_output
        self.n_channels = n_channels
        self.n_weights = n_weights
        self.n_filters = n_filters
        self.filter_size = filter_size
        self.n_hlayers = len(n_weights)
        self.n_conv = len(n_filters)
        self.n_modes = n_modes
        self.drate = drate
        self.maxpool = maxpool

        network_weights = self._create_weights()
        self.weights = network_weights

        self.nonlinearity = tf.nn.relu
        self.nonlinearity_mean = tf.clip_by_value

    def _calc_z_mean_and_sigma(self,x):
        with tf.name_scope("VI_encoder_r1"):
 
            # Reshape input to a 3D tensor - single channel
            if self.n_conv is not None:
                conv_pool = tf.reshape(x, shape=[-1, self.n_input, 1, self.n_channels])
                for i in range(self.n_conv):
                    weight_name = 'w_conv_' + str(i)
                    bias_name = 'b_conv_' + str(i)
                    bnorm_name = 'bn_conv_' + str(i)
                    conv_pre1 = tf.add(tf.nn.conv2d(conv_pool, self.weights['VI_encoder_r1'][weight_name + '1'],strides=1,padding='SAME'),self.weights['VI_encoder_r1'][bias_name + '1'])
                    conv_post = self.nonlinearity(conv_pre1)
                    conv_batchnorm = tf.nn.batch_normalization(conv_post,tf.Variable(tf.zeros([1,1,self.n_filters[i]], dtype=tf.float32)),tf.Variable(tf.ones([1,1,self.n_filters[i]], dtype=tf.float32)),None,None,1e-6,name=bnorm_name)
                    conv_pool = tf.nn.max_pool2d(conv_batchnorm,ksize=[self.maxpool[i],1],strides=[self.maxpool[i],1],padding='SAME')

                fc = tf.reshape(conv_pool, [-1, int(self.n_input*self.n_filters[-1]/(np.prod(self.maxpool)))])

            else:
                fc = tf.reshape(x,[-1,self.n_input*self.n_channels])
           
            hidden_dropout = fc
            for i in range(self.n_hlayers):
                weight_name = 'w_hidden_' + str(i)
                bias_name = 'b_hidden' + str(i)
                bnorm_name = 'bn_hidden' + str(i)
                hidden_pre = tf.add(tf.matmul(hidden_dropout, self.weights['VI_encoder_r1'][weight_name]), self.weights['VI_encoder_r1'][bias_name])
                hidden_post = self.nonlinearity(hidden_pre)
                hidden_batchnorm = tf.nn.batch_normalization(hidden_post,tf.Variable(tf.zeros([self.n_weights[i]], dtype=tf.float32)),tf.Variable(tf.ones([self.n_weights[i]], dtype=tf.float32)),None,None,1e-6,name=bnorm_name)
                hidden_dropout = tf.layers.dropout(hidden_batchnorm,rate=self.drate)
            loc = tf.add(tf.matmul(hidden_dropout, self.weights['VI_encoder_r1']['w_loc']), self.weights['VI_encoder_r1']['b_loc'])
            scale_diag = tf.add(tf.matmul(hidden_dropout, self.weights['VI_encoder_r1']['w_scale_diag']), self.weights['VI_encoder_r1']['b_scale_diag'])
            #scale_tri = tf.add(tf.matmul(hidden_dropout, self.weights['VICI_encoder']['w_scale_tri']), self.weights['VICI_encoder']['b_scale_tri'])
            weight = tf.add(tf.matmul(hidden_dropout, self.weights['VI_encoder_r1']['w_weight']), self.weights['VI_encoder_r1']['b_weight']) 

            # make +ve definite covariance matrix
            #A_rs = tf.reshape(scale_tri,(-1,int((self.n_output*(self.n_output+1))/2)))
            #A_posdiag = tf.linalg.set_diag(A_rs,SMALL_CONSTANT + tf.exp(tf.linalg.diag_part(A_rs)))
            #A = tfp.math.fill_triangular(A_posdiag, upper=False, name=None)
            #cov = tf.linalg.matmul(A, A, transpose_a=False, transpose_b=True)

            tf.summary.histogram('loc', loc)
            tf.summary.histogram('scale_diag', scale_diag)
            #tf.summary.histogram('scale_tri', scale_tri)
            tf.summary.histogram('weight', weight)
            return tf.reshape(loc,(-1,self.n_modes,self.n_output)), tf.reshape(scale_diag,(-1,self.n_modes,self.n_output)), tf.reshape(weight,(-1,self.n_modes))    
            #return tf.reshape(loc,(-1,self.n_modes,self.n_output)), tf.reshape(A,(-1,self.n_modes,self.n_output,self.n_output)), tf.reshape(weight,(-1,self.n_modes))

    def _create_weights(self):
        all_weights = collections.OrderedDict()
        with tf.variable_scope("VI_ENC_r1"):            
            all_weights['VI_encoder_r1'] = collections.OrderedDict()

            if self.n_conv is not None:
                dummy = self.n_channels
                for i in range(self.n_conv):
                    weight_name = 'w_conv_' + str(i)
                    bias_name = 'b_conv_' + str(i)
                    all_weights['VI_encoder_r1'][weight_name + '1'] = tf.Variable(tf.reshape(vae_utils.xavier_init(self.filter_size[i], dummy*self.n_filters[i]),[self.filter_size[i], 1, dummy, self.n_filters[i]]), dtype=tf.float32)
                    all_weights['VI_encoder_r1'][bias_name + '1'] = tf.Variable(tf.zeros([self.n_filters[i]], dtype=tf.float32))
                    #all_weights['VICI_encoder'][weight_name + '2'] = tf.Variable(tf.reshape(vae_utils.xavier_init(self.filter_size*2, dummy*self.n_filters),[self.filter_size*2, dummy, self.n_filters]), dtype=tf.float32)
                    #all_weights['VICI_encoder'][bias_name + '2'] = tf.Variable(tf.zeros([self.n_filters], dtype=tf.float32))
                    #all_weights['VICI_encoder'][weight_name + '3'] = tf.Variable(tf.reshape(vae_utils.xavier_init(self.filter_size*4, dummy*self.n_filters),[self.filter_size*4, dummy, self.n_filters]), dtype=tf.float32)
                    #all_weights['VICI_encoder'][bias_name + '3'] = tf.Variable(tf.zeros([self.n_filters], dtype=tf.float32))
                    tf.summary.histogram(weight_name + '1', all_weights['VI_encoder_r1'][weight_name + '1'])
                    tf.summary.histogram(bias_name + '1', all_weights['VI_encoder_r1'][bias_name + '1'])
                    #tf.summary.histogram(weight_name + '2', all_weights['VICI_encoder'][weight_name + '2'])
                    #tf.summary.histogram(bias_name + '2', all_weights['VICI_encoder'][bias_name + '2'])
                    #tf.summary.histogram(weight_name + '3', all_weights['VICI_encoder'][weight_name + '3'])
                    #tf.summary.histogram(bias_name + '3', all_weights['VICI_encoder'][bias_name + '3'])
                    dummy = self.n_filters[i]

                fc_input_size = int(self.n_input*self.n_filters[-1]/(np.prod(self.maxpool)))
            else:
                fc_input_size = self.n_input*self.n_channels

            for i in range(self.n_hlayers):
                weight_name = 'w_hidden_' + str(i)
                bias_name = 'b_hidden' + str(i)
                all_weights['VI_encoder_r1'][weight_name] = tf.Variable(vae_utils.xavier_init(fc_input_size, self.n_weights[i]), dtype=tf.float32)
                all_weights['VI_encoder_r1'][bias_name] = tf.Variable(tf.zeros([self.n_weights[i]], dtype=tf.float32))
                tf.summary.histogram(weight_name, all_weights['VI_encoder_r1'][weight_name])
                tf.summary.histogram(bias_name, all_weights['VI_encoder_r1'][bias_name])
                fc_input_size = self.n_weights[i]
            all_weights['VI_encoder_r1']['w_loc'] = tf.Variable(vae_utils.xavier_init(self.n_weights[-1], self.n_output*self.n_modes),dtype=tf.float32)
            all_weights['VI_encoder_r1']['b_loc'] = tf.Variable(tf.zeros([self.n_output*self.n_modes], dtype=tf.float32), dtype=tf.float32)
            tf.summary.histogram('w_loc', all_weights['VI_encoder_r1']['w_loc'])
            tf.summary.histogram('b_loc', all_weights['VI_encoder_r1']['b_loc'])
            all_weights['VI_encoder_r1']['w_scale_diag'] = tf.Variable(vae_utils.xavier_init(self.n_weights[-1], self.n_output*self.n_modes),dtype=tf.float32)
            all_weights['VI_encoder_r1']['b_scale_diag'] = tf.Variable(tf.zeros([self.n_output*self.n_modes], dtype=tf.float32), dtype=tf.float32)
            tf.summary.histogram('w_scale', all_weights['VI_encoder_r1']['w_scale_diag'])
            tf.summary.histogram('b_scale', all_weights['VI_encoder_r1']['b_scale_diag'])
            #all_weights['VICI_encoder']['w_scale_tri'] = tf.Variable(vae_utils.xavier_init(self.n_weights[-1], int((self.n_output*(self.n_output+1)*self.n_modes)/2)),dtype=tf.float32)
            #all_weights['VICI_encoder']['b_scale_tri'] = tf.Variable(tf.zeros([int((self.n_output*(self.n_output+1)*self.n_modes)/2)], dtype=tf.float32), dtype=tf.float32)
            #tf.summary.histogram('w_scale_tri', all_weights['VICI_encoder']['w_scale_tri'])
            #tf.summary.histogram('b_scale_tri', all_weights['VICI_encoder']['b_scale_tri'])
            all_weights['VI_encoder_r1']['w_weight'] = tf.Variable(vae_utils.xavier_init(self.n_weights[-1], self.n_modes),dtype=tf.float32)
            all_weights['VI_encoder_r1']['b_weight'] = tf.Variable(tf.zeros([self.n_modes], dtype=tf.float32), dtype=tf.float32)
            tf.summary.histogram('w_weight', all_weights['VI_encoder_r1']['w_weight'])
            tf.summary.histogram('b_weight', all_weights['VI_encoder_r1']['b_weight'])

            all_weights['prior_param'] = collections.OrderedDict()
        
        return all_weights



import collections

#import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import numpy as np
import math as m

from . import vae_utils

# based on implementation here:
# https://github.com/tensorflow/models/blob/master/autoencoder/autoencoder_models/VariationalAutoencoder.py

class VariationalAutoencoder(object):

    def __init__(self, name, n_input1=3, n_input2=256, n_output=4, n_channels=3, n_weights=2048, drate=0.2, n_filters=8, filter_size=8, maxpool=4):
        
        self.n_input1 = n_input1
        self.n_input2 = n_input2
        self.n_output = n_output
        self.n_channels = n_channels
        self.n_weights = n_weights
        self.n_hlayers = len(n_weights)
        self.n_conv = len(n_filters)
        self.drate = drate
        self.n_filters = n_filters
        self.filter_size = filter_size
        self.maxpool = maxpool

        network_weights = self._create_weights()
        self.weights = network_weights
        self.nonlinearity = tf.nn.relu

    def _calc_z_mean_and_sigma(self,x,y):
        with tf.name_scope("VI_encoder_q"):

            # Reshape input to a 3D tensor - single channel
            if self.n_conv is not None:
                conv_pool = tf.reshape(y, shape=[-1, self.n_input2, 1, self.n_channels])
                for i in range(self.n_conv):
                    weight_name = 'w_conv_' + str(i)
                    bias_name = 'b_conv_' + str(i)
                    bnorm_name = 'b_conv_' + str(i)
                    conv_pre1 = tf.add(tf.nn.conv2d(conv_pool, self.weights['VI_encoder_q'][weight_name + '1'],strides=1,padding='SAME'),self.weights['VI_encoder_q'][bias_name + '1'])
                    #conv_pre2 = tf.add(tf.nn.conv1d(conv_pool, self.weights['VICI_VAE_encoder'][weight_name + '2'],stride=1,padding='SAME'),self.weights['VICI_VAE_encoder'][bias_name + '2'])
                    #conv_pre3 = tf.add(tf.nn.conv1d(conv_pool, self.weights['VICI_VAE_encoder'][weight_name + '3'],stride=1,padding='SAME'),self.weights['VICI_VAE_encoder'][bias_name + '3'])
                    #conv_post = self.nonlinearity(tf.concat([conv_pre1,conv_pre2,conv_pre3],axis=2))
                    conv_post = self.nonlinearity(conv_pre1)
                    conv_batchnorm = tf.nn.batch_normalization(conv_post,tf.Variable(tf.zeros([1,1,self.n_filters[i]], dtype=tf.float32)),tf.Variable(tf.ones([1,1,self.n_filters[i]], dtype=tf.float32)),None,None,1e-6,name=bnorm_name)
                    #conv_dropout = tf.nn.dropout(conv_batchnorm,rate=self.drate)
                    conv_pool = tf.nn.max_pool2d(conv_batchnorm,ksize=[self.maxpool[i],1],strides=[self.maxpool[i],1],padding='SAME')

                fc = tf.concat([x,tf.reshape(conv_pool, [-1, int(self.n_input2*self.n_filters[-1]/(np.prod(self.maxpool)))])],axis=1)

            else:
                fc = tf.concat([x,y],axis=1)

            hidden_dropout = fc
            for i in range(self.n_hlayers):
                weight_name = 'w_hidden_' + str(i)
                bias_name = 'b_hidden' + str(i)
                hidden_pre = tf.add(tf.matmul(hidden_dropout, self.weights['VI_encoder_q'][weight_name]), self.weights['VI_encoder_q'][bias_name])
                hidden_post = self.nonlinearity(hidden_pre)
                hidden_batchnorm = tf.nn.batch_normalization(hidden_post,tf.Variable(tf.zeros([self.n_weights[i]], dtype=tf.float32)),tf.Variable(tf.ones([self.n_weights[i]], dtype=tf.float32)),None,None,1e-6,name=bnorm_name)
                hidden_dropout = tf.layers.dropout(hidden_batchnorm,rate=self.drate)
            loc = tf.add(tf.matmul(hidden_dropout, self.weights['VI_encoder_q']['w_loc']), self.weights['VI_encoder_q']['b_loc'])
            scale = tf.add(tf.matmul(hidden_dropout, self.weights['VI_encoder_q']['w_scale']), self.weights['VI_encoder_q']['b_scale'])

            tf.summary.histogram('loc', loc)
            tf.summary.histogram('scale', scale)
            return loc, scale


    def _sample_from_gaussian_dist(self, num_rows, num_cols, mean, log_sigma_sq):
        with tf.name_scope("sample_in_z_space"):
            eps = tf.random.normal([num_rows, num_cols], 0, 1., dtype=tf.float32)
            sample = tf.add(mean, tf.multiply(tf.sqrt(tf.exp(log_sigma_sq)), eps))
        return sample

    def _create_weights(self):
        all_weights = collections.OrderedDict()
        with tf.variable_scope("VI_ENC_q"):
            # Encoder
            all_weights['VI_encoder_q'] = collections.OrderedDict()
            
            if self.n_conv is not None:
                dummy = self.n_channels
                for i in range(self.n_conv):
                    weight_name = 'w_conv_' + str(i)
                    bias_name = 'b_conv_' + str(i)
                    all_weights['VI_encoder_q'][weight_name + '1'] = tf.Variable(tf.reshape(vae_utils.xavier_init(self.filter_size[i], dummy*self.n_filters[i]),[self.filter_size[i], 1, dummy, self.n_filters[i]]), dtype=tf.float32)
                    all_weights['VI_encoder_q'][bias_name + '1'] = tf.Variable(tf.zeros([self.n_filters[i]], dtype=tf.float32))
                    #all_weights['VICI_VAE_encoder'][weight_name + '2'] = tf.Variable(tf.reshape(vae_utils.xavier_init(self.filter_size*2, dummy*self.n_filters),[self.filter_size*2, dummy, self.n_filters]), dtype=tf.float32)
                    #all_weights['VICI_VAE_encoder'][bias_name + '2'] = tf.Variable(tf.zeros([self.n_filters], dtype=tf.float32))
                    #all_weights['VICI_VAE_encoder'][weight_name + '3'] = tf.Variable(tf.reshape(vae_utils.xavier_init(self.filter_size*4, dummy*self.n_filters),[self.filter_size*4, dummy, self.n_filters]), dtype=tf.float32)
                    #all_weights['VICI_VAE_encoder'][bias_name + '3'] = tf.Variable(tf.zeros([self.n_filters], dtype=tf.float32))
                    tf.summary.histogram(weight_name + '1', all_weights['VI_encoder_q'][weight_name + '1'])
                    tf.summary.histogram(bias_name + '1', all_weights['VI_encoder_q'][bias_name + '1'])
                    #tf.summary.histogram(weight_name + '2', all_weights['VICI_VAE_encoder'][weight_name + '2'])
                    #tf.summary.histogram(bias_name + '2', all_weights['VICI_VAE_encoder'][bias_name + '2'])
                    #tf.summary.histogram(weight_name + '3', all_weights['VICI_VAE_encoder'][weight_name + '3'])
                    #tf.summary.histogram(bias_name + '3', all_weights['VICI_VAE_encoder'][bias_name + '3'])
                    dummy = self.n_filters[i]

                fc_input_size = self.n_input1 + int(self.n_input2*self.n_filters[-1]/(np.prod(self.maxpool)))
            else:
                fc_input_size = self.n_input1 + self.n_input2*self.n_channels

            for i in range(self.n_hlayers):
                weight_name = 'w_hidden_' + str(i)
                bias_name = 'b_hidden' + str(i)
                all_weights['VI_encoder_q'][weight_name] = tf.Variable(vae_utils.xavier_init(fc_input_size, self.n_weights[i]), dtype=tf.float32)
                all_weights['VI_encoder_q'][bias_name] = tf.Variable(tf.zeros([self.n_weights[i]], dtype=tf.float32))
                tf.summary.histogram(weight_name, all_weights['VI_encoder_q'][weight_name])
                tf.summary.histogram(bias_name, all_weights['VI_encoder_q'][bias_name])
                fc_input_size = self.n_weights[i]
            all_weights['VI_encoder_q']['w_loc'] = tf.Variable(vae_utils.xavier_init(self.n_weights[-1], self.n_output),dtype=tf.float32)
            all_weights['VI_encoder_q']['b_loc'] = tf.Variable(tf.zeros([self.n_output], dtype=tf.float32), dtype=tf.float32)
            tf.summary.histogram('w_loc', all_weights['VI_encoder_q']['w_loc'])
            tf.summary.histogram('b_loc', all_weights['VI_encoder_q']['b_loc'])           
            all_weights['VI_encoder_q']['w_scale'] = tf.Variable(vae_utils.xavier_init(self.n_weights[-1], self.n_output),dtype=tf.float32)
            all_weights['VI_encoder_q']['b_scale'] = tf.Variable(tf.zeros([self.n_output], dtype=tf.float32), dtype=tf.float32)
            tf.summary.histogram('w_scale', all_weights['VI_encoder_q']['w_scale'])
            tf.summary.histogram('b_scale', all_weights['VI_encoder_q']['b_scale'])

            all_weights['prior_param'] = collections.OrderedDict()
        
        return all_weights



""" Script used to generate run parameters json files
"""

import numpy as np
import os
import json

# Source parameter values to use if chosen to be fixed
fixed_vals = {'mass_1':50.0,
        '__definition__mass_1':'m1 fixed value',
        'mass_2':50.0,
        '__definition__mass_2':'m2 fixed value',
        'mc':None,
        '__definition__mc': 'chirp mass fixed value',
        'geocent_time':0.0,
        '__definition__geocent_time': 'time of coalescence fixed value',
        'phase':0.0,
        '__definition__phase': 'phase fixed value',
        'ra':1.375,
        '__definition__ra': 'right ascension fixed value',
        'dec':-1.2108,
        '__definition__dec': 'declination fixed value',
        'psi':0.0,
        '__definition__psi': 'psi fixed value',
        'theta_jn':0.0,
        '__definition__theta_jn': 'inclination angle fixed value',
        'luminosity_distance':2000.0,
        '__definition__luminosity_distance': 'luminosity distance fixed value',
        'a_1':0.0,
        '__definition__a1': 'a1 fixed value',
        'a_2':0.0,
        '__definition__a2': 'a2 fixed value',
	    'tilt_1':0.0,
        '__definition__tilt1': 'tilt_1 fixed value',
	    'tilt_2':0.0,
        '__definition__tilt_2': 'tilt_2 fixed value',
        'phi_12':0.0,
        '__definition__phi_12': 'phi_12 fixed value',
        'phi_jl':0.0,
        '__definition__phi_jl': 'phi_jl fixed value'}                                                

# Prior bounds on source parameters
bounds = {'mass_1_min':35.0, 'mass_1_max':80.0,
        '__definition__mass_1': 'mass 1 range',
        'mass_2_min':35.0, 'mass_2_max':80.0,
        '__definition__mass_2': 'mass 2 range',
        'M_min':70.0, 'M_max':160.0,
        '__definition__M': 'total mass range',
        'geocent_time_min':0.15,'geocent_time_max':0.35,
        '__definition__geocent_time': 'time of coalescence range',
        'phase_min':0.0, 'phase_max':2.0*np.pi,
        '__definition__phase': 'phase range',
        'ra_min':0.0, 'ra_max':2.0*np.pi,
        '__definition__ra': 'right ascension range',
        'dec_min':-0.5*np.pi, 'dec_max':0.5*np.pi,
        '__definition__dec': 'declination range',
        'psi_min':0.0, 'psi_max':np.pi,
        '__definition__psi': 'psi range',
        'theta_jn_min':0.0, 'theta_jn_max':np.pi,
        '__definition__theta_jn': 'inclination angle range',
        'a_1_min':0.0, 'a_1_max':0.8,
        '__definition__a_1': 'a1 range',
        'a_2_min':0.0, 'a_2_max':0.8,
        '__definition__a_2': 'a2 range',
        'tilt_1_min':0.0, 'tilt_1_max':np.pi,
        '__definition__tilt1': 'tilt1 range',
        'tilt_2_min':0.0, 'tilt_2_max':np.pi,
        '__definition__tilt2': 'tilt2 range',
        'phi_12_min':0.0, 'phi_12_max':2.0*np.pi,
        '__definition__phi12': 'phi12 range',
        'phi_jl_min':0.0, 'phi_jl_max':2.0*np.pi,
        '__definition_phijl': 'phijl range',
        'luminosity_distance_min':1000.0, 'luminosity_distance_max':3000.0,
        '__definition__luminosity_distance': 'luminosity distance range'}

# arbitrary value for normalization of timeseries (Don't change this)
y_normscale = 36.0

##########################
# Main tunable variables
##########################
ndata = 256                                                                     
det=['H1','L1','V1']                                                            
psd_files=[] 
rand_pars = ['mass_1','mass_2','luminosity_distance','geocent_time','phase',
                 'theta_jn','psi','ra','dec']#,'a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl','ra','dec']                                   
inf_pars=['mass_1','mass_2','luminosity_distance','geocent_time','theta_jn','ra','dec']#,'a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl','ra','dec'] 
batch_size = 64                                                                 
weight_init = 'xavier'                                                            
n_modes=7                                                                      
initial_training_rate=1e-4                                                     
batch_norm=True                                                                

# FYI, each item in lists below correspond to each layer in networks (i.e. first item first layer)
# pool size and pool stride should be same number in each layer
n_filters_r1 = [33, 33, 33, 33, 33]                                                        
n_filters_r2 = [33, 33, 33]                                                        
n_filters_q = [33, 33, 33]                                                         
filter_size_r1 = [5,8,11,10,10]                                                         
filter_size_r2 = [5,8,11]                                                        
filter_size_q = [5,8,11]                                                          
drate = 0.5                                                                    
maxpool_r1 = [1,2,1,2,1]                                                             
conv_strides_r1 = [1,1,1,1,1]                                                        
pool_strides_r1 = [1,2,1,2,1]                                                        
maxpool_r2 = [1,2,1]                                                             
conv_strides_r2 = [1,1,1]                                                        
pool_strides_r2 = [1,2,1]                                                        
maxpool_q = [1,2,1]                                                              
conv_strides_q = [1,1,1]                                                         
pool_strides_q = [1,2,1]                                                         
n_fc = 2048                                                                      
z_dimension=10                                                                  
n_weights_r1 = [n_fc,n_fc,n_fc]                                                     
n_weights_r2 = [n_fc,n_fc,n_fc]                                                     
n_weights_q = [n_fc,n_fc,n_fc]                                                      
##########################
# Main tunable variables
##########################

#############################
# optional tunable variables
#############################
run_label = 'demo_%ddet_%dpar_%dHz' % (len(det),len(rand_pars),ndata) 
bilby_results_label = 'bilby_results'                                             
r = 1                                                                           
pe_test_num = 256                                                               
tot_dataset_size = int(1e7)                                                     
tset_split = int(1e3)                                                           
save_interval = int(1e3)                                                        
num_iterations=int(1e6)+1                                                       
#ref_geocent_time=1126259642.5                                                   
ref_geocent_time=1126259642.5 #+ ((86400.0/2.0))
load_chunk_size = 1e3                                                           
samplers=['vitamin','dynesty']                                                  

# Directory variables
plot_dir="./results/%s" % run_label  
#train_set_dir='/home/hunter.gabbard/CBC/VItamin/training_sets_second_sub_3det_9par_256Hz/tset_tot-10000000_split-1000/'#'./training_sets_%ddet_%dpar_%dHz/tset_tot-%d_split-%d' % (len(det),len(rand_pars),ndata,tot_dataset_size,tset_split) 
#test_set_dir='/home/hunter.gabbard/CBC/public_VItamin/testing_directory/test_sets/all_4_samplers/test_waveforms'#'./test_sets/%s/test_waveforms' % bilby_results_label                                                           
#pe_dir='/home/hunter.gabbard/CBC/public_VItamin/testing_directory/test_sets/all_4_samplers/test'#'./test_sets/%s/test' % bilby_results_label                                                                            
#test_set_dir='./daniel_williams_BNU_waveforms/hunter-mdc'
train_set_dir='./training_sets_%ddet_%dpar_%dHz/tset_tot-%d_split-%d' % (len(det),len(rand_pars),ndata,tot_dataset_size,tset_split)
test_set_dir='./test_sets/%s/test_waveforms' % bilby_results_label
pe_dir='./test_sets/%s/test' % bilby_results_label
#############################
# optional tunable variables

# Function for getting list of parameters that need to be fed into the models
def get_params():

    # Define dictionary to store values used in rest of code 
    params = dict(
        make_corner_plots = True,                                               
        __definition__make_corner_plots='if True, make corner plots',
        make_kl_plot = True,                                                    
        __definition__make_kl_plot='If True, go through kl plotting function',
        make_pp_plot = True,                                                    
        __definition__make_pp_plot='If True, go through pp plotting function',
        make_loss_plot = False,                                                 
        __definition__make_loss_plot='If True, generate loss plot from previous plot data',
        Make_sky_plot=False,                                                    
        __definition__make_sky_plot='If True, generate sky plots on corner plots',
        hyperparam_optim = False,                                               
        __definition__hyperparam_optim='optimize hyperparameters for model during training using gaussian process minimization',
        resume_training=False,                                                  
        __definition___resume_training='if True, resume training of a model from saved checkpoint',
        load_by_chunks = True,                                                  
        __definition__load_by_chunks='if True, load training samples by a predefined chunk size rather than all at once',	
        ramp = True,                                                            
        __definition__ramp='# if true, apply linear ramp to KL loss',
        print_values=True,                                                      
        __definition__print_values='# optionally print loss values every report interval',
        by_channel = True,                                                      
        __definition__by_channel='if True, do convolutions as seperate 1-D channels, if False, stack training samples as 2-D images (n_detectors,(duration*sampling_frequency))',
        load_plot_data=False,                                                   
        __definition__load_plot_data='Use plotting data which has already been generated',
        doPE = False,                                                            
        __definition__doPE='if True then do bilby PE when generating new testing samples',
        gpu_num=0,                                                              
        __definition__gpu_num='gpu number run is running on',
        ndata = ndata,                                                          
        __definition__ndata='sampling frequency',
        run_label=run_label,                                                    
        __definition__run_label='label of run',
        bilby_results_label=bilby_results_label,                                
        __definition__bilby_result_label='label given to bilby results directory',
        tot_dataset_size = tot_dataset_size, 
        __definition__tot_dataset_size='total number of training samples available to use',                                   
        tset_split = tset_split,
        __definition__tset_split='number of training samples in each training data file',                                                
        plot_dir=plot_dir,
        __definition__plot_dir='output directory to save results plots',

        # Gaussian Process automated hyperparameter tunning variables
        hyperparam_optim_stop = int(1.5e6),                                     
        __definition__hyperparam_optim_stop='stopping iteration of hyperparameter optimizer per call (ideally 1.5 million)', 
        hyperparam_n_call = 30,                                                 
        __definition__hyperparam_n_call='number of hyperparameter optimization calls (ideally 30)',
        
        load_chunk_size = load_chunk_size,
        __definition__load_chunk_size='Number of training samples to load in at a time.',                                      
        load_iteration = int((load_chunk_size * 25)/batch_size),                
        __definition__load_iteration='How often to load another chunk of training samples',
        weight_init = weight_init,
        __definition__weight_init='[xavier,VarianceScaling,Orthogonal]. Network model weight initialization (default is xavier)',                                           
        n_samples = 10000,                                                        
        __definition__n_samples='number of posterior samples to save per reconstruction upon inference (default 3000)',
        num_iterations=num_iterations,                                              
        __definition__num_iterations='total number of iterations before ending training of model',
        initial_training_rate=initial_training_rate,
        __definition__initial_training_rate='initial training rate for ADAM optimiser inference model (inverse reconstruction)',                         
        batch_size=batch_size,
        __definition__batch_size=' Number training samples shown to neural network per iteration',                                                  
        batch_norm=batch_norm,
        __definition__batch_norm='if true, do batch normalization in all layers of neural network',                                                  
        report_interval=500,                                                    
        __definition__report_interval='interval at which to save objective function values and optionally print info during training',
        n_modes=n_modes,
        __definition__n_modes='number of modes in Gaussian mixture model (ideal 7, but may go higher/lower)',                                                     

        # FYI, each item in lists below correspond to each layer in networks (i.e. first item first layer)
        # pool size and pool stride should be same number in each layer
        n_filters_r1 = n_filters_r1,
        __definition__n_filters_r1='number of convolutional filters to use in r1 network (must be divisible by 3)',                                         
        n_filters_r2 = n_filters_r2,
        __definition__n_filters_r2='number of convolutional filters to use in r2 network (must be divisible by 3)',                                         
        n_filters_q = n_filters_q,
        __definition__n_filters_q='number of convolutional filters to use in q network  (must be divisible by 3)',                                           
        filter_size_r1 = filter_size_r1,  
        __definition__filter_size_r1='size of convolutional fitlers in r1 network',                                   
        filter_size_r2 = filter_size_r2,   
        __definition__filter_size_r2='size of convolutional filters in r2 network',                                  
        filter_size_q = filter_size_q,
        __definition__filter_size_q='size of convolutional filters in q network',                                       
        drate = drate,  
        __definition__drate='dropout rate to use in fully-connected layers',                                                     
        maxpool_r1 = maxpool_r1,                                            
        __definition__maxpool_r1='size of maxpooling to use in r1 network', 
        conv_strides_r1 = conv_strides_r1,
        __definition__conv_strides_r1='size of convolutional stride to use in r1 network',                                   
        pool_strides_r1 = pool_strides_r1,                                   
        __definition_pool_strides_r1='size of max pool stride to use in r1 network',
        maxpool_r2 = maxpool_r2,       
        __definition_maxpool_r2='size of max pooling to use in r2 network',                                      
        conv_strides_r2 = conv_strides_r2,
        __definition__conv_strides_r2='size of convolutional stride in r2 network',                                   
        pool_strides_r2 = pool_strides_r2,                                   
        __definition__pool_strides_r2='size of max pool stride in r2 network',
        maxpool_q = maxpool_q,  
        __definition__maxpool_q='size of max pooling to use in q network',                                             
        conv_strides_q = conv_strides_q,  
        __definition__conv_strides_q='size of convolutional stride to use in q network',                                   
        pool_strides_q = pool_strides_q,                                     
        __definition__pool_strides_q='size of max pool stride to use in q network',
        ramp_start = 1e4,                                                       
        __definition__ramp_start='starting iteration of KL divergence ramp',
        ramp_end = 1e5,                                                  
        __definition__ramp_end='ending iteration of KL divergence ramp',
        save_interval=save_interval, 
        __definition__save_interval='number of iterations to save model',                                           
        plot_interval=save_interval,                                     
        __definition__plot_interval='number of iterations to plot validation results corner plots',       
        z_dimension=z_dimension,                                              
        __definition__z_dimension='number of latent space dimensions of model',
        n_weights_r1 = n_weights_r1,                                         
        __definition__n_weights_r1='number fully-connected neurons in layers of encoders and decoders in the r1 model',
        n_weights_r2 = n_weights_r2,
        __definition__n_weights_r2='number fully-connected neurons in layers of encoders and decoders in the r2 model',                                         
        n_weights_q = n_weights_q,                                           
        __definition__n_weights_q='number fully-connected neurons in layers of encoders and decoders in the q model',
        duration = 1.0,                                                         
        __definition__duration='length of training/validation/test sample time series in seconds',
        r = r,     
        __definition__r='number of GW timeseries to use for testing.',                                                             
        rand_pars=rand_pars,                                                    
        __definition__rand_pars='parameters to randomize (those not listed here are fixed otherwise)',
        corner_labels = {'mass_1': '$m_{1}\,(\mathrm{M}_{\odot})$',
                         'mass_2': '$m_{2}\,(\mathrm{M}_{\odot})$',
                         'luminosity_distance': '$d_{\mathrm{L}}\,(\mathrm{Mpc})$',
                         'geocent_time': '$t_{0}\,(\mathrm{seconds})$',
                         'phase': '${\phi}$',
                         'theta_jn': '$\Theta_{jn}\,(\mathrm{rad})$',
                         'psi': '${\psi}$',
                         'ra': r'${\alpha}\,(\mathrm{rad})$',
                         'dec': '${\delta}\,(\mathrm{rad})$',
                         'a_1': '${a_1}$',
                         'a_2': '${a_2}$',
                         'tilt_1': '${\Theta_{1}}$',
                         'tilt_2': '${\Theta_{2}}$',
                         'phi_12': '${\phi_{12}}$',
                         'phi_jl': '${\phi_{jl}}$'},

        __definition__corner_labels='latex source parameter labels for plotting',
        ref_geocent_time=ref_geocent_time,                                      
        __definition__ref_geocent_time='reference gps time (not advised to change this)',
        training_data_seed=43,                                                 
        __definition__training_data_seed='tensorflow training random seed number',
        testing_data_seed=44,                                                   
        __definition__testing_data_seed=' tensorflow testing random seed number',
        inf_pars=inf_pars,        
        __definition__inf_pars='parameters to infer',                                              
        train_set_dir=train_set_dir,
        __definition__train_set_dir='location of training set directory',
        test_set_dir=test_set_dir,
        __definition__test_set_dir='location of test set directory waveforms',
        pe_dir=pe_dir,
        __definition__pe_dir='location of test set directory Bayesian PE samples. Set to None if you do not want to use posterior samples when training.',
        KL_cycles = 1,                                                          
        __definition__KL_cycles=' number of cycles to repeat for the KL approximation',
        samplers=samplers,       
        __definition__samplers='Samplers to use when comparing ML results (vitamin is ML approach) dynesty,ptemcee,cpnest,emcee',                                               
        figure_sampler_names = ['VItamin','Dynesty'],
        __definition__figure_sampler_names='matplotlib figure sampler labels (e.g. [ptemcee,CPNest,emcee])',
        y_normscale = y_normscale,
        __definition__y_normscale='arbitrary normalization factor on all time series waveforms (helps convergence in training)',
        boost_pars=['ra','dec'],
        __definition__boost_pars='parameters to boost during training (by default this is off)',
        gauss_pars=['luminosity_distance','geocent_time','theta_jn','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'],         
        __definition__gauss_pars='parameters that require a truncated gaussian distribution',
        vonmise_pars=['phase','psi'],                                        
        __definition__vonmises_pars='parameters that get wrapped on the 1D parameter', 
        sky_pars=['ra','dec'],                                              
        __definition__sky_pars='sky parameters',
        det=det,
        __definition__det='LIGO detectors to perform analysis on (default is 3detector H1,L1,V1)',
        psd_files=psd_files,
        __definition__psd_files='User may specficy their own psd files for each detector. Must be bilby compatible .txt files. If specifying your own files, do so for each detector. If list is empty, default Bilby PSD values will be used for each detector. ',
        weighted_pars=None,
        __definition__weighted_pars='set to None if not using, parameters to weight during training',
        weighted_pars_factor=1,                       
        __definition__weighted_pars_factor='factor by which to weight weighted parameters',
    )
    return params


# Save training/test parameters of run if files do not already exist
params=get_params()

# Make directory containing params files
try:
    os.mkdir('params_files')
except FileExistsError:
    print("directory 'params_files' already exists")

# Generate params files
if not os.path.isfile('./params_files/params.json'):
    
    # save params json file
    with open('params_files/params.json', 'w') as fp:
        json.dump(params, fp, sort_keys=False, indent=4)

if not os.path.isfile('./params_files/bounds.json'):

    # save bounds json file
    with open('params_files/bounds.json', 'w') as fp:
        json.dump(bounds, fp, sort_keys=False, indent=4)

if not os.path.isfile('./params_files/fixed_vals.json'):

    # save fixed vals json file
    with open('params_files/fixed_vals.json', 'w') as fp:
        json.dump(fixed_vals, fp, sort_keys=False, indent=4)



#!/usr/bin/env python

from setuptools import setup
import subprocess
import sys
import os

# check that python version is 3.5 or above
python_version = sys.version_info
if python_version < (3, 0):
    sys.exit("Python < 3.0 is not supported, aborting setup")
print("Confirmed Python version {}.{}.{} >= 3.0.0".format(*python_version[:3]))


def write_version_file(version):
    """ Writes a file with version information to be used at run time

    Parameters
    ----------
    version: str
        A string containing the current version information

    Returns
    -------
    version_file: str
        A path to the version file

    """
    try:
        git_log = subprocess.check_output(
            ['git', 'log', '-1', '--pretty=%h %ai']).decode('utf-8')
        git_diff = (subprocess.check_output(['git', 'diff', '.']) +
                    subprocess.check_output(
                        ['git', 'diff', '--cached', '.'])).decode('utf-8')
        if git_diff == '':
            git_status = '(CLEAN) ' + git_log
        else:
            git_status = '(UNCLEAN) ' + git_log
    except Exception as e:
        print("Unable to obtain git version information, exception: {}"
              .format(e))
        git_status = 'release'

    version_file = '.version'
    if os.path.isfile(version_file) is False:
        with open('vitamin_b/' + version_file, 'w+') as f:
            f.write('{}: {}'.format(version, git_status))

    return version_file


def get_long_description():
    """ Finds the README and reads in the description """
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, 'README.md')) as f:
        long_description = f.read()
    return long_description


# get version info from __init__.py
def readfile(filename):
    with open(filename) as fp:
        filecontents = fp.read()
    return filecontents


VERSION = '0.2.11'
version_file = write_version_file(VERSION)
long_description = get_long_description()


setup(
    name='vitamin_b',
    version='0.2.11',    
    description='A user-friendly machine learning Bayesian inference library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/hagabbar/vitamin_b',
    author='Hunter Gabbard, Chris Messenger, Ik Siong Heng, Francesco Tonolini, Roderick Murray-Smith',
    author_email='h.gabbard.1@research.gla.ac.uk',
    license='GNU General Public License v3 (GPLv3)',
    packages=['vitamin_b','vitamin_b.models','vitamin_b.models.neural_networks',
              'vitamin_b.params_files'],
    package_dir={'vitamin_b': 'vitamin_b'},
    package_data={'vitamin_b': ['params_files/*.json'],
                  'vitamin_b': [version_file]},

    python_requires='>=3.5', 
    install_requires=['universal-divergence',
                      'absl-py',
                      'asn1crypto',
                      'astor',
                      'astroplan',
                      'astropy',
                      'astropy-healpix',
                      'astroquery',
                      'beautifulsoup4',
                      'bilby',
                      'cachetools',
                      'certifi',
                      'cffi',
                      'chardet',
                      'cloudpickle',
                      'corner',
                      'cpnest',
                      'cryptography',
                      'cycler',
                      'Cython',
                      'decorator',
                      'deepdish',
                      'dill',
                      'dqsegdb2',
                      'dynesty',
                      'emcee',
                      'entrypoints',
                      'future',
                      'gast',
                      'google-auth',
                      'google-auth-oauthlib',
                      'google-pasta',
                      'grpcio',
                      'gwdatafind',
                      'gwosc',
                      'gwpy',
                      'h5py',
                      'healpy',
                      'html5lib',
                      'idna',
                      'jeepney',
                      'joblib',
                      'Keras',
                      'Keras-Applications',
                      'Keras-Preprocessing',
                      'keyring',
                      'kiwisolver',
                      'lalsuite',
                      'ligo-gracedb',
                      'ligo-segments',
                      'ligo.skymap',
                      'ligotimegps',
                      'lscsoft-glue',
                      'Markdown',
                      'matplotlib',
                      'mock',
                      'networkx',
                      'numexpr',
                      'numpy',
                      'oauthlib',
                      'opt-einsum',
                      'pandas',
                      'patsy',
                      'Pillow',
                      'protobuf',
                      'ptemcee',
                      'pyaml',
                      'pyasn1',
                      'pyasn1-modules',
                      'pycparser',
                      'pymc3',
                      'pyOpenSSL',
                      'pyparsing',
                      'python-dateutil',
                      'python-ligo-lw',
                      'pytz',
                      'PyYAML',
                      'reproject',
                      'requests',
                      'requests-oauthlib',
                      'rsa',
                      'scikit-learn',
                      'scikit-optimize',
                      'scipy',
                      'seaborn',
                      'SecretStorage',
                      'six',
                      'soupsieve',
                      'tables',
                      'tensorboard',
                      'tensorflow==2.1.0',
                      'tensorflow-estimator',
                      'tensorflow-probability==0.9.0',
                      'termcolor',
                      'tqdm',
                      'urllib3',
                      'webencodings',
                      'Werkzeug',
                      'wrapt',
                      ],

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',  
        'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 3.6',
    ],
)




import collections

#import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import numpy as np
import math as m

from . import vae_utils

# based on implementation here:
# https://github.com/tensorflow/models/blob/master/autoencoder/autoencoder_models/VariationalAutoencoder.py

class VariationalAutoencoder(object):

    def __init__(self, name, wrap_mask, nowrap_mask, m1_mask, m2_mask, sky_mask, n_input1=4, n_input2=256, n_output=3, n_channels=3, n_weights=2048, drate=0.2, n_filters=8, filter_size=8, maxpool=4):
        
        self.n_input1 = n_input1                    # actually the output size
        self.n_input2 = n_input2                    # actually the output size
        self.n_output = n_output                    # the input data size
        self.n_channels = n_channels                # the number of channels/detectors
        self.n_weights = n_weights                  # the number of weights were layer
        self.n_hlayers = len(n_weights)             # the number of fully connected layers
        self.n_conv = len(n_filters)                # the number of convolutional layers
        self.n_filters = n_filters                  # the number of filters in each conv layer
        self.filter_size = filter_size              # the filter sizes in each conv layer
        self.maxpool = maxpool                      # the max pooling sizes in each conv layer
        self.name = name                            # the name of the network
        self.drate = drate                          # dropout rate
        self.wrap_mask = wrap_mask                  # mask identifying wrapped indices
        self.nowrap_mask = nowrap_mask              # mask identifying non-wrapped indices
        self.m1_mask = m1_mask                      # the mask identifying the m1 parameter
        self.m2_mask = m2_mask                      # the mask identifying the m2 parameter
        self.sky_mask = sky_mask                    # the mask identifying the sky (RA,dec) parameters
        self.nonlinear_loc_nowrap = tf.sigmoid      # activation for non-wrapped location params
        self.nonlinear_loc_wrap = tf.sigmoid        # activation for wrapped location params
        self.nonlinear_loc_m1 = tf.sigmoid          # activation for mass params
        self.nonlinear_loc_m2 = tf.sigmoid          # activation for mass params
        self.nonlinear_loc_sky = tf.identity        # activation for sky params
        self.nonlinear_scale_nowrap = tf.identity   # activation for non-wrapped scale params
        self.nonlinear_scale_wrap = tf.nn.relu      # activation for wrapped scale params
        self.nonlinear_scale_m1 = tf.nn.relu        # activation for mass params
        self.nonlinear_scale_m2 = tf.nn.relu        # activation for mass params  
        self.nonlinear_scale_sky = tf.nn.relu       # activation for sky params
        self.nonlinearity = tf.nn.relu              # activation between hidden layers

        network_weights = self._create_weights()
        self.weights = network_weights

    def calc_reconstruction(self, z, y):
        with tf.name_scope("VI_decoder_r2"):

            # Reshape input to a 3D tensor - single channel
            if self.n_conv is not None:
                conv_pool = tf.reshape(y, shape=[-1, self.n_input2, 1, self.n_channels])
                for i in range(self.n_conv):            
                    weight_name = 'w_conv_' + str(i)
                    bias_name = 'b_conv_' + str(i)
                    bnorm_name = 'bn_conv_' + str(i)
                    conv_pre1 = tf.add(tf.nn.conv2d(conv_pool, self.weights['VI_decoder_r2'][weight_name + '1'],strides=1,padding='SAME'),self.weights['VI_decoder_r2'][bias_name + '1'])
                    conv_post = self.nonlinearity(conv_pre1)
                    conv_batchnorm = tf.nn.batch_normalization(conv_post,tf.Variable(tf.zeros([1,1,self.n_filters[i]], dtype=tf.float32)),tf.Variable(tf.ones([1,1,self.n_filters[i]], dtype=tf.float32)),None,None,1e-6,name=bnorm_name)
                    conv_pool = tf.nn.max_pool2d(conv_batchnorm,ksize=[self.maxpool[i],1],strides=[self.maxpool[i],1],padding='SAME')

                fc = tf.concat([z,tf.reshape(conv_pool, [-1, int(self.n_input2*self.n_filters[-1]/(np.prod(self.maxpool)))])],axis=1)            

            else:
                fc = tf.concat([z,y],axis=1)

            hidden_dropout = fc
            for i in range(self.n_hlayers):
                weight_name = 'w_hidden_' + str(i)
                bias_name = 'b_hidden' + str(i)
                bnorm_name = 'bn_hidden' + str(i)
                hidden_pre = tf.add(tf.matmul(hidden_dropout, self.weights['VI_decoder_r2'][weight_name]), self.weights['VI_decoder_r2'][bias_name])
                hidden_post = self.nonlinearity(hidden_pre)
                hidden_batchnorm = tf.nn.batch_normalization(hidden_post,tf.Variable(tf.zeros([self.n_weights[i]], dtype=tf.float32)),tf.Variable(tf.ones([self.n_weights[i]], dtype=tf.float32)),None,None,1e-6,name=bnorm_name)
                hidden_dropout = tf.layers.dropout(hidden_batchnorm,rate=self.drate)
            loc_all = tf.add(tf.matmul(hidden_dropout, self.weights['VI_decoder_r2']['w_loc']), self.weights['VI_decoder_r2']['b_loc'])
            scale_all = tf.add(tf.matmul(hidden_dropout, self.weights['VI_decoder_r2']['w_scale']), self.weights['VI_decoder_r2']['b_scale'])

            # split up the output into non-wrapped and wrapped params and apply appropriate activation
            loc_nowrap = self.nonlinear_loc_nowrap(tf.boolean_mask(loc_all,self.nowrap_mask + [False],axis=1))   # add an extra null element to the mask
            scale_nowrap = self.nonlinear_scale_nowrap(tf.boolean_mask(scale_all,self.nowrap_mask[:-1],axis=1))  # ignore last element because scale_all is 1 shorter
            loc_m1 = self.nonlinear_loc_m1(tf.boolean_mask(loc_all,self.m1_mask + [False],axis=1))             # add an extra null element to the mask
            scale_m1 = -1.0*self.nonlinear_scale_m1(tf.boolean_mask(scale_all,self.m1_mask[:-1],axis=1))      # ignore last element because scale_all is 1 shorter
            loc_m2 = self.nonlinear_loc_m2(tf.boolean_mask(loc_all,self.m2_mask + [False],axis=1))            # add an extra null element to the mask
            scale_m2 = -1.0*self.nonlinear_scale_m2(tf.boolean_mask(scale_all,self.m2_mask[:-1],axis=1))    # ignore last element because scale_all is 1 shorter
            loc_wrap = self.nonlinear_loc_wrap(tf.boolean_mask(loc_all,self.wrap_mask + [False],axis=1))    # add an extra null element to the mask 
            scale_wrap = -1.0*self.nonlinear_scale_wrap(tf.boolean_mask(scale_all,self.wrap_mask[:-1],axis=1))  # ignore last element because scale_all is 1 shorter
            loc_sky = self.nonlinear_loc_sky(tf.boolean_mask(loc_all,self.sky_mask + [True],axis=1))        # add an extra element to the mask for the 3rd sky parameter
            scale_sky = -1.0*self.nonlinear_scale_sky(tf.boolean_mask(scale_all,self.sky_mask[:-1],axis=1))    # ignore last element because scale_all is 1 shorter
            return loc_nowrap, scale_nowrap, loc_wrap, scale_wrap, loc_m1, scale_m1, loc_m2, scale_m2, loc_sky, scale_sky   

    def _create_weights(self):
        all_weights = collections.OrderedDict()

        # Decoder
        with tf.variable_scope("VI_DEC"):
            all_weights['VI_decoder_r2'] = collections.OrderedDict()
            
            if self.n_conv is not None:
                dummy = self.n_channels
                for i in range(self.n_conv):
                    weight_name = 'w_conv_' + str(i)
                    bias_name = 'b_conv_' + str(i)
                    all_weights['VI_decoder_r2'][weight_name + '1'] = tf.Variable(tf.reshape(vae_utils.xavier_init(self.filter_size[i], dummy*self.n_filters[i]),[self.filter_size[i], 1, dummy, self.n_filters[i]]), dtype=tf.float32)
                    all_weights['VI_decoder_r2'][bias_name + '1'] = tf.Variable(tf.zeros([self.n_filters[i]], dtype=tf.float32))
                    tf.summary.histogram(weight_name + '1', all_weights['VI_decoder_r2'][weight_name + '1'])
                    tf.summary.histogram(bias_name + '1', all_weights['VI_decoder_r2'][bias_name + '1'])
                    dummy = self.n_filters[i]

                fc_input_size = self.n_input1 + int(self.n_input2*self.n_filters[-1]/(np.prod(self.maxpool)))
            else:
                fc_input_size = self.n_input1 + self.n_input2*self.n_channels

            for i in range(self.n_hlayers):
                weight_name = 'w_hidden_' + str(i)
                bias_name = 'b_hidden' + str(i)
                all_weights['VI_decoder_r2'][weight_name] = tf.Variable(vae_utils.xavier_init(fc_input_size, self.n_weights[i]), dtype=tf.float32)
                all_weights['VI_decoder_r2'][bias_name] = tf.Variable(tf.zeros([self.n_weights[i]], dtype=tf.float32))
                tf.summary.histogram(weight_name, all_weights['VI_decoder_r2'][weight_name])
                tf.summary.histogram(bias_name, all_weights['VI_decoder_r2'][bias_name])
                fc_input_size = self.n_weights[i]
            all_weights['VI_decoder_r2']['w_loc'] = tf.Variable(vae_utils.xavier_init(self.n_weights[-1], self.n_output+1),dtype=tf.float32)  # +1 for extra sky param
            all_weights['VI_decoder_r2']['b_loc'] = tf.Variable(tf.zeros([self.n_output+1], dtype=tf.float32), dtype=tf.float32) # +1 for extra sky param
            tf.summary.histogram('w_loc', all_weights['VI_decoder_r2']['w_loc'])
            tf.summary.histogram('b_loc', all_weights['VI_decoder_r2']['b_loc'])
            all_weights['VI_decoder_r2']['w_scale'] = tf.Variable(vae_utils.xavier_init(self.n_weights[-1], self.n_output-1),dtype=tf.float32) # # -1 for common concentration par in VMF dist
            all_weights['VI_decoder_r2']['b_scale'] = tf.Variable(tf.zeros([self.n_output-1], dtype=tf.float32), dtype=tf.float32)  # -1 for common concentration par in VMF dist
            tf.summary.histogram('w_scale', all_weights['VI_decoder_r2']['w_scale'])
            tf.summary.histogram('b_scale', all_weights['VI_decoder_r2']['b_scale'])
            
            all_weights['prior_param'] = collections.OrderedDict()
        
        return all_weights
